import datasets

_DESCRIPTION = """\
Paraphrase datasets extracted from paired sentences in Wikipedia and Simple Wikipedia
"""

_CITATION = """\
  Improving Text Simplification Language Modeling Using Unsimplified Text Data. David Kauchak (2013). In Proceedings of ACL.
"""

_DATA_URL = {
    "v1": "https://cs.pomona.edu/~dkauchak/simplification/data.v1/data.v1.tar.gz",
    "v2": "https://cs.pomona.edu/~dkauchak/simplification/data.v2/sentence-aligned.v2.tar.gz",
}


class SimpleWikiConfig(datasets.BuilderConfig):
    def __init__(self, version, **kwargs):
        """BuilderConfig for SimpleWiki.
        Args:
            version: version of simple wiki dataset (v1 or v2)
            **kwargs: keyword arguments forwarded to super.
        """
        name = f"SimpleWiki.{version}"

        description = (f"Paraphrase dataset from Simple Wikipedia, version {version}")
        super(SimpleWikiConfig, self).__init__(
            name=name,
            description=description,
            version=datasets.Version("0.0.1", ""),
            **kwargs,
        )

        self.ds_version = version


class SimpleWiki(datasets.GeneratorBasedBuilder):
    """FLoRes machine translation dataset."""

    BUILDER_CONFIGS = [
        SimpleWikiConfig("v1"),
        SimpleWikiConfig("v2"),
    ]

    def _info(self):
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {"source": datasets.Value("string"),
                 "target": datasets.Value("string")}
            ),
            supervised_keys=None,
            homepage="https://cs.pomona.edu/~dkauchak/simplification/",
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        path = dl_manager.download_and_extract(_DATA_URL[self.config.ds_version])

        if self.config.ds_version == 'v1':
            return [
                datasets.SplitGenerator(name=datasets.Split.TRAIN, gen_kwargs={
                    "source": f"{path}/data.v1/wiki.unsimplified",
                    "target": f"{path}/data.v1/wiki.simple",
                }),
            ]
        elif self.config.ds_version == 'v2':
            return [
                datasets.SplitGenerator(name=datasets.Split.TRAIN, gen_kwargs={
                    "source": f"{path}/sentence-aligned.v2/normal.aligned",
                    "target": f"{path}/sentence-aligned.v2/simple.aligned",
                }),
            ]

    def _generate_examples(self, source, target):
        """This function returns the examples in the raw (text) form."""

        def extract_sentences(f):
            sentences = []
            lines = f.readlines()
            for l in lines:
                l = l.rstrip().split('\t')
                sentences.append(l[-1])
            print(f"extracted {len(sentences)}")
            return sentences

        with open(source) as f_normal, open(target) as f_simple:
            source_sentences = extract_sentences(f_normal)
            target_sentences = extract_sentences(f_simple)

        assert len(target_sentences) == len(source_sentences), "Sizes do not match: %d vs %d" % (
            len(source_sentences),
            len(target_sentences)
        )

        for idx, (l1, l2) in enumerate(zip(source_sentences, target_sentences)):
            result = {"source": l1, "target": l2}
            # Make sure that both translations are non-empty.
            if all(result.values()):
                yield idx, result
