from typing import Optional

import pytorch_lightning as pl
import torch.nn.functional as F
from torch.utils import data

from .prepclass import *

DS_PROCESS_METHOD_MAP = {
    'SimpleWiki.v1': SimpleWikiPrep,
    'Quora': QuoraPrep,
    'WikiText-103': WikiTextPrep
}


class Src2TgtLenDataModule(pl.LightningDataModule):
    def __init__(self, dataset='SimpleWiki.v1',
                 vocab='glove6B', num_unk=1,
                 maxlen=0, mode='text',
                 bs=256):
        super(Src2TgtLenDataModule, self).__init__()
        self.dataset = dataset
        self.vocab = vocab
        self.vocabname = f'{vocab}_unk{num_unk}'
        self.maxlen = maxlen
        self.mode = mode
        self.num_unk = num_unk
        self.preper = DS_PROCESS_METHOD_MAP[dataset](self.dataset, self.vocab, self.num_unk, self.maxlen, self.mode)
        if self.mode == 'text':
            self.collate_fn = self.text_collate_fn
        elif self.mode == 'unsup':
            self.collate_fn = self.syn_collate_fn

        self.bs = bs

    def prepare_data(self):
        self.preper.prepare()

    def setup(self, stage: Optional[str] = None):
        # self.vocabpt = torch.load(f'Data/vocab/{self.vocabname}.pt')
        train_wrapper = self.preper.get_wrapper(self.mode, 'train')
        test_wrapper = self.preper.get_wrapper(self.mode, 'test')
        if stage in (None, 'fit'):
            ds = self.preper.load('train').train_test_split(test_size=0.1)
            self.trainds = ds['train']
            self.trainds.set_format(type='torch')
            self.valds = ds['test']
            self.valds.set_format(type='torch')
            if train_wrapper is not None:
                self.trainds = train_wrapper(self.trainds)
            if test_wrapper is not None:
                self.valds = test_wrapper(self.valds)

        if stage in (None, 'test') or stage in (None, 'predict'):
            self.testds = self.preper.load('test')
            self.testds.set_format(type='torch')
            if test_wrapper is not None:
                self.testds = test_wrapper(self.testds)

    def text_collate_fn(self, batch):
        # first stack lengths
        src_len = self.stack_tensors([d['src_len'] for d in batch])
        tgt_len = self.stack_tensors([d['tgt_len'] for d in batch])
        if self.maxlen != 0:
            src_len = torch.clamp(src_len, max=self.maxlen)
            tgt_len = torch.clamp(tgt_len, max=self.maxlen)
        src_max = max(src_len).item()
        tgt_max = max(tgt_len).item()
        src_seq = self.stack_tensors([self.pad1d(d['source'], src_max) for d in batch])
        tgt_seq = self.stack_tensors([self.pad1d(d['target'], tgt_max) for d in batch])
        return {'src_seq': src_seq, 'src_len': src_len,
                'tgt_seq': tgt_seq, 'tgt_len': tgt_len}

    def syn_collate_fn(self, batch):
        # {'source': tok_src, 'target': tok_tgt,
        #  'src_len': src_len, 'tgt_len': tgt_len,
        #  'tgt_h': tgt_h_seq, 'tgt_r': tgt_r_seq, 'tgt_dep': tgt_dep,}
        # first stack lengths
        src_len = self.stack_tensors([d['src_len'] for d in batch])
        tgt_len = self.stack_tensors([d['tgt_len'] for d in batch])
        if self.maxlen != 0:
            src_len = torch.clamp(src_len, max=self.maxlen)
            tgt_len = torch.clamp(tgt_len, max=self.maxlen)
        src_max = max(src_len).item()
        tgt_max = max(tgt_len).item()
        src_seq = self.stack_tensors([self.pad1d(d['source'], src_max) for d in batch])
        tgt_seq = self.stack_tensors([self.pad1d(d['target'], tgt_max) for d in batch])
        tgt_h = self.stack_tensors([self.pad1d(d['tgt_h'], tgt_max - 1) for d in batch])
        tgt_r = self.stack_tensors([self.pad1d(d['tgt_r'], tgt_max - 1) for d in batch])
        tgt_dep = self.stack_tensors([self.pad1d(d['tgt_dep'], tgt_max - 1) for d in batch])
        return {'src_seq': src_seq, 'src_len': src_len,
                'tgt_seq': tgt_seq, 'tgt_len': tgt_len,
                'tgt_h': tgt_h, 'tgt_r': tgt_r, 'tgt_dep': tgt_dep}

    def train_dataloader(self):
        return data.DataLoader(self.trainds, batch_size=self.bs, shuffle=True, collate_fn=self.collate_fn,
                               num_workers=8)

    def val_dataloader(self):
        return data.DataLoader(self.valds, batch_size=self.bs, collate_fn=self.collate_fn, num_workers=4)

    def test_dataloader(self):
        return data.DataLoader(self.testds, batch_size=self.bs, collate_fn=self.collate_fn)

    def predict_dataloader(self):
        return self.test_dataloader()

    @staticmethod
    def stack_tensors(bat):
        elem = bat[0]
        out = None
        if torch.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum(x.numel() for x in bat)
            storage = elem.storage()._new_shared(numel)
            out = elem.new(storage)
        return torch.stack(bat, 0, out=out)

    @staticmethod
    def pad1d(s, l):
        if len(s) > l:
            return s[:l]
        else:
            return F.pad(s, [0, l - s.size(0)], 'constant', 0)
