import os
import random
import subprocess

import datasets
import stanza
import torch
from datasets import load_dataset, load_from_disk
from .wrappers import Para2SelfEncWrapper, TextSelfEncWrapper, SynUnsupWrapper
from ..utils import tree_depth


class DatasetPreper:
    def __init__(self, dataset='SimpleWiki.v1',
                 vocab='glove6B', num_unk=1,
                 maxlen=0, mode='text'):
        self.dataset = dataset
        self.vocab = f'{vocab}_unk{num_unk}'
        self.maxlen = maxlen
        self.mode = mode
        self.num_unk = num_unk
        self.unk_trick = False
        self.unk_ids = ['<unk>']
        if self.num_unk > 1:
            self.unk_trick = True
            self.unk_ids = [f'<unk{i}>' for i in range(self.num_unk)]

        self.tokened = False
        self.token_ds = f'{self.dataset}_tokonly'
        if os.path.exists(f'Data/ds/{self.token_ds}'):
            self.tokened = True

        self.prepared = False
        name_mode = 'syn' if self.mode == 'unsup' else self.mode
        self.prep_ds = f'{self.dataset}_{self.vocab}_{name_mode}_{self.maxlen}'
        if os.path.exists(f'Data/ds/{self.prep_ds}'):
            if self.mode != 'text':
                ds = load_from_disk(f'Data/ds/{self.prep_ds}')
                feats = ds['train'].features
                if 'tgt_dep' not in feats and 'src_dep' not in feats:
                    print('Upgrade syn mode dataset:')
                    ds = ds.map(self.mapfunc_syn_upgrade, batched=True, batch_size=50)
                    ds.save_to_disk(f'Data/ds/{self.prep_ds}')
            self.prepared = True

        self.vocabpt = None
        self.text_cols = None
        self.nlp = None
        self.rel_vocabpt = None
        # self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

    def load_and_token(self):
        raise NotImplementedError

    def process_splits(self, ds):
        return ds['train'].train_test_split(test_size=0.1)

    def load(self, split):
        if not self.prepared:
            self.prepare()
        return load_from_disk(f'Data/ds/{self.prep_ds}')[split]

    def prepare(self):
        if self.prepared:
            return
        # first check tokenize
        if not self.tokened:
            ds = self.load_and_token()
            ds = self.process_splits(ds)
            ds.save_to_disk(f'Data/ds/{self.token_ds}')
        else:
            ds = load_from_disk(f'Data/ds/{self.token_ds}')
        self.vocabpt = torch.load(f'Data/vocab/{self.vocab}.pt')

        # then generate mode dataset
        if self.mode == 'text':
            prep_ds = ds.map(self.mapfunc_text_mode, batched=True, batch_size=50)
        elif self.mode == 'unsup':
            self.nlp = stanza.Pipeline('en', processors='tokenize, pos, lemma, depparse',
                                       tokenize_pretokenized=True)
            self.rel_vocabpt = torch.load(f'Data/vocab/relvocab.pt')
            # semvec = f'Data/semvec/{self.dataset}_glove.42B.300d.txt'
            # self.svloss = LossModel(self.num_unk, semvec).to(self.device)
            #
            # print(f'Synvec loss model loaded on {self.svloss.unk_head_map.device}')
            prep_ds = ds.map(self.mapfunc_syn_mode, batched=True, batch_size=100)
            self.nlp = None
            self.rel_vocabpt = None
        else:
            print('Unknown mode of datasets')
            exit(1)
        if self.maxlen != 0 and self.text_cols is not None:
            prep_ds = prep_ds.filter(lambda ex:
                                     max([1 if len(ex[col]) > self.maxlen else 0 for col in self.text_cols]) == 0
                                     )
        prep_ds.save_to_disk(f'Data/ds/{self.prep_ds}')
        self.vocabpt = None
        self.prepared = True

    def mapfunc_text_mode(self, samples):
        tok_src = []
        tok_tgt = []
        src_len = []
        tgt_len = []
        for i in range(len(samples['source'])):
            t_tok_src = self.sentence2id(samples['source'][i])
            t_tok_tgt = self.sentence2id(samples['target'][i])
            tok_src.append(t_tok_src)
            tok_tgt.append(t_tok_tgt)
            src_len.append(len(t_tok_src))
            tgt_len.append(len(t_tok_tgt))
        return {'source': tok_src, 'target': tok_tgt,
                'src_len': src_len, 'tgt_len': tgt_len}

    def mapfunc_syn_mode(self, samples):
        tok_src = []
        tok_tgt = []
        src_len = []
        tgt_len = []
        src_h_seq = []
        src_r_seq = []
        tgt_h_seq = []
        tgt_r_seq = []
        src_dep = []
        tgt_dep = []
        for i in range(len(samples['source'])):
            t_tok_src = self.sentence2id(samples['source'][i])
            t_tok_tgt = self.sentence2id(samples['target'][i], eos=False)
            tok_src.append(t_tok_src)
            tok_tgt.append(t_tok_tgt)
            src_len.append(len(t_tok_src))
            tgt_len.append(len(t_tok_tgt))
            t_src_h_seq, t_src_r_seq = self.sentence2synseq(samples['source'][i])
            t_tgt_h_seq, t_tgt_r_seq = self.sentence2synseq(samples['target'][i])
            src_h_seq.append(t_src_h_seq)
            src_r_seq.append(t_src_r_seq)
            tgt_h_seq.append(t_tgt_h_seq)
            tgt_r_seq.append(t_tgt_r_seq)
            src_dep.append(tree_depth(t_src_h_seq))
            tgt_dep.append(tree_depth(t_tgt_h_seq))
        res = {'source': tok_src, 'target': tok_tgt,
               'src_len': src_len, 'tgt_len': tgt_len,
               'src_h': src_h_seq, 'src_r': src_r_seq, 'src_dep': src_dep,
               'tgt_h': tgt_h_seq, 'tgt_r': tgt_r_seq, 'tgt_dep': tgt_dep}
        return res

    @staticmethod
    def mapfunc_syn_upgrade(samples):
        src_dep = []
        tgt_dep = []
        for i in range(len(samples['source'])):
            src_dep.append(tree_depth(samples['src_h'][i]))
            tgt_dep.append(tree_depth(samples['tgt_h'][i]))
        samples.update({'src_dep': src_dep, 'tgt_dep': tgt_dep})
        return samples

    def sentence2synseq(self, sentence):
        parse_res = self.nlp([sentence])
        # head_seq = [0]
        # rel_seq = [self.rel_vocabpt['root']]
        head_seq = []
        rel_seq = []
        for w in parse_res.sentences[0].words:
            head_seq.append(w.head)
            rel_seq.append(self.rel_vocabpt[w.deprel.split(':')[0]])
        return head_seq, rel_seq

    def sentence2id(self, sentence, eos=True):
        if self.vocabpt is None:
            self.vocabpt = torch.load(f'Data/vocab/{self.vocab}.pt')
        tokened = [self.vocabpt['<bos>']]
        unk_map = {}
        for w in sentence:
            if w in self.vocabpt:
                tokened.append(self.vocabpt[w])
                continue
            if self.unk_trick:
                if w not in unk_map:
                    unk_map[w] = random.choice(self.unk_ids)
                tokened.append(self.vocabpt[unk_map[w]])
            else:
                tokened.append(self.vocabpt[w])
        if eos:
            tokened.append(self.vocabpt['<eos>'])
        return tokened

    @staticmethod
    def get_wrapper(mode='text', split='train'):
        if mode == 'text':
            return Para2SelfEncWrapper
        elif mode == 'unsup' and split == 'train':
            return SynUnsupWrapper
        else:
            return None


class SimpleWikiPrep(DatasetPreper):
    def __init__(self, *args, **kwargs):
        super(SimpleWikiPrep, self).__init__(*args, **kwargs)
        print(args, kwargs)
        self.text_cols = ['source', 'target']

    def load_and_token(self):
        ds = load_dataset('Data/SimpleWiki/simplewiki.py', name=self.dataset)
        ds = ds.map(self.mapfunc_tok, batched=True, batch_size=50)
        return ds

    def mapfunc_tok(self, samples):
        tok_src = []
        tok_tgt = []
        for i in range(len(samples['source'])):
            t_tok_src = [self.normalize(w) for w in samples['source'][i].split(' ')]
            t_tok_tgt = [self.normalize(w) for w in samples['target'][i].split(' ')]
            tok_src.append(t_tok_src)
            tok_tgt.append(t_tok_tgt)
        return {'source': tok_src, 'target': tok_tgt}

    def normalize(self, word):
        if word == "''" or word == "``":
            word = '"'
        if word == "`":
            word = "'"
        return word.lower()


class QuoraPrep(DatasetPreper):
    def __init__(self, *args, **kwargs):
        super(QuoraPrep, self).__init__(*args, **kwargs)
        print(args, kwargs)
        self.text_cols = ['source', 'target']

    def load_and_token(self):
        ds = load_dataset('quora')
        self.nlp = stanza.Pipeline('en', processors='tokenize')
        ds = ds.filter(lambda x: x['is_duplicate'])
        ds = ds.map(self.mapfunc_conv_tok, batched=True, batch_size=50,
                    remove_columns=['is_duplicate', 'questions'])
        self.nlp = None
        return ds

    def mapfunc_conv_tok(self, samples):
        src_q = []
        tgt_q = []
        for i in range(len(samples['is_duplicate'])):
            tmp_text = samples['questions'][i]['text']
            tmp_src_q = []
            tmp_src_sent = self.nlp(tmp_text[0])
            for s in tmp_src_sent.sentences:
                for w in s.words:
                    tmp_src_q.append(w.text.lower())
            tmp_tgt_q = []
            tmp_tgt_sent = self.nlp(tmp_text[1])
            for s in tmp_tgt_sent.sentences:
                for w in s.words:
                    tmp_tgt_q.append(w.text.lower())
            src_q.append(tmp_src_q)
            tgt_q.append(tmp_tgt_q)
        return {'source': src_q, 'target': tgt_q}


class WikiTextPrep(DatasetPreper):
    def __init__(self, *args, **kwargs):
        super(WikiTextPrep, self).__init__(*args, **kwargs)
        print(args, kwargs)
        self.blacklist = {
            '@-@', '@.@', '@,@', '==', '=', ''
        }
        self.text_cols = ['text']

    def load_and_token(self):
        ds = load_dataset('wikitext', 'wikitext-103-raw-v1')
        self.nlp = stanza.Pipeline('en', processors='tokenize')
        ds = ds.map(self.mapfunc_tok, batched=True, batch_size=50)
        ds = ds.filter(lambda ex: len(ex['text']) >= 5)
        self.nlp = None
        return ds

    def mapfunc_tok(self, samples):
        tok_txt = []
        for i in range(len(samples['text'])):
            t_tok_txt = list(filter(lambda x: x not in self.blacklist,
                                    [w.lower() for w in samples['text'][i].rstrip().split(' ')]))
            t_tok_txt = " ".join(t_tok_txt)
            res = self.nlp(t_tok_txt)
            for s in res.sentences:
                t_tok_txt = []
                for w in s.words:
                    t_tok_txt.append(w.text.lower())
                tok_txt.append(t_tok_txt)
        return {'text': tok_txt}

    def mapfunc_text_mode(self, samples):
        tokens = []
        length = []
        for text in samples['text']:
            t_tok = self.sentence2id(text)
            tokens.append(t_tok)
            length.append(len(t_tok))
        return {'text': tokens, 'length': length}

    @staticmethod
    def get_wrapper(mode='text', split='train'):
        if mode == 'text':
            return TextSelfEncWrapper
        else:
            return None

    def process_splits(self, ds):
        ds['train'] = datasets.concatenate_datasets([ds['train'], ds['validation']])
        ds.pop('validation')
        return ds
