from collections import OrderedDict, Counter

import datasets
import torch
from datasets import load_from_disk
from torchtext.vocab import vocab
from tqdm import trange

from pysyngen.utils import read_vectors


def build_spl_token_list(num_unk=1):
    spl_tok_list = ['<pad>', '<bos>', '<eos>']
    if num_unk == 1:
        spl_tok_list.append('<unk>')
    else:
        for i in range(num_unk):
            spl_tok_list.append(f'<unk{i}>')
    return spl_tok_list


def build_vocab_from_word_list(path, special_token_list=None, name=None):
    wordlist = special_token_list.copy() if special_token_list is not None else []
    f = open(path, 'r')
    for line in f:
        wordlist.append(line.rstrip().split()[0])
    voc = vocab(OrderedDict([(token, 1) for token in wordlist]))
    if special_token_list is not None:
        # default unknown token is 3
        voc.set_default_index(3)
    if name is None:
        return voc
    else:
        if special_token_list is None:
            name = name
        else:
            name = name + f'_unk{len(special_token_list) - 3}'
        torch.save(voc, f'Data/vocab/{name}.pt')


def build_vocab_from_dataset(dsname, columns=('source', 'target'), sp_token=None,
                             save_name=None,
                             vocab_size=-1, vocab_freq=1,
                             we_vocab=None, we_name='we_cus'):
    if sp_token is not None:
        od = OrderedDict([(token, 1) for token in sp_token])
    else:
        od = OrderedDict()
    ds = load_from_disk(f'Data/ds/{dsname}_tokonly')
    counter = Counter([])
    for split in ds.keys():
        for sid in trange(ds[split].num_rows, desc=split):
            for col in columns:
                if we_vocab is None:
                    counter.update(ds[split][sid][col])
                else:
                    counter.update(filter(lambda w: w in we_vocab, ds[split][sid][col]))
    num_tokens = sum(counter.values())
    print(f'Processed corpus with {num_tokens} tokens')
    word_counts = sorted(counter.items(), key=lambda x: x[1], reverse=True)
    ori_count = len(word_counts)
    if isinstance(vocab_freq, int) and vocab_freq > 1:
        print(f'Filtering words with count gt {vocab_freq} equivalent to freq gt {vocab_freq / num_tokens:.8f}')
        word_counts = list(filter(lambda x: x[1] >= vocab_freq, word_counts))
    elif isinstance(vocab_freq, float):
        vf = int(num_tokens * vocab_freq + 1)
        print(f'Filtering words with freq gt {vocab_freq} equivalent to count gt {vf}')
        word_counts = list(filter(lambda x: x[1] >= vf, word_counts))
    if vocab_size != -1:
        word_counts = word_counts[:vocab_size]
    od.update(OrderedDict(word_counts))
    voc = vocab(od)
    if sp_token is not None:
        # default unknown token is 3
        voc.set_default_index(3)
    if we_vocab is not None:
        name = f'{dsname}_{we_name}'
    else:
        name = dsname
    if save_name is not None:
        name = save_name
    if sp_token is not None:
        name += f'_unk{len(sp_token) - 3}'

    torch.save(voc, f'Data/vocab/{name}.pt')
    print(ori_count, len(voc))
    return voc


def build_semvec_from_vocab(osemvec, nvocab, savepath, num_unk=1, dim=300):
    _, iw, wi = read_vectors(osemvec, dim, cache=False)
    with open(osemvec) as f_ovec:
        ovec = f_ovec.readlines()
    n_vectors = []
    num_spl_token = num_unk + 3
    for wid in range(num_spl_token, len(nvocab)):
        w = nvocab.lookup_token(wid)
        o_wid = wi[w]
        n_vectors.append(ovec[o_wid])
    print(len(nvocab), len(nvocab) - num_spl_token, len(n_vectors))
    with open(savepath, 'w') as f_nvec:
        f_nvec.writelines(n_vectors)


def update_vocab_with_spl(voc, new_spl, save_name):
    # first find old specials
    current_unk_id = voc['<unk>']
    for i in range(1, len(voc) - 3):
        tmp_unk_id = voc[f'<unk{i}>']
        if tmp_unk_id <= current_unk_id:
            break
        else:
            current_unk_id = tmp_unk_id
    start_id = current_unk_id + 1
    # construct a vocab based on special tokens
    fake_count = [(token, 1) for token in new_spl]
    fake_count.extend([(w, 1) for w in voc.get_itos()[start_id:]])
    od = OrderedDict(fake_count)
    new_voc = vocab(od)
    new_voc.set_default_index(3)
    save_name += f'_unk{len(new_spl) - 3}'
    torch.save(new_voc, f'Data/vocab/{save_name}.pt')


class VocabMerger:
    def __init__(self, main_vocab, ref_vocab):
        self.mvoc = main_vocab
        self.rvoc = ref_vocab
        self.diff_counts = {}

    def count_diff(self, ds_name, columns=('source', 'target'), count_thresh=None):
        ds = datasets.load_from_disk(f'Data/ds/{ds_name}_tokonly')
        self.diff_counts[ds_name] = Counter([])
        cnt_rows = 0
        for split in ds.keys():
            for sid in trange(ds[split].num_rows, desc=split):
                for col in columns:
                    cnt_rows += 1
                    self.diff_counts[ds_name].update(filter(
                        lambda w: w in self.rvoc and w not in self.mvoc,
                        ds[split][sid][col]))
        word_counts = sorted(self.diff_counts[ds_name].items(), key=lambda x: x[1], reverse=True)
        self.diff_counts[ds_name] = word_counts
        print(f'{ds_name} has {len(word_counts)} new words:')
        for i, w in enumerate(word_counts[:15]):
            print(f'{w[0]}({w[1]}/{w[1] / cnt_rows:.4f})', end=' ')
            if i % 5 == 0 and i != 0:
                print()
        if count_thresh is not None:
            for i, w in enumerate(word_counts):
                if w[1] <= count_thresh:
                    print(f'{i + 1} words with counts higher than {count_thresh}')
                    break
        return

    def shrink_counts(self, ds_name, count_thresh, size_thresh, preview=True):
        old_length = len(self.diff_counts[ds_name])
        new_counts = list(filter(lambda x: x[1] >= count_thresh, self.diff_counts[ds_name]))[:size_thresh]
        print(f'{ds_name} diff list shrunk to {len(new_counts)} from {old_length}')
        if not preview:
            self.diff_counts[ds_name] = new_counts

    def merge_all(self):
        new_voc = self.mvoc  # .clone() how the f**k to make sure a copy
        new_words_set = set()
        for k in self.diff_counts:
            words = [w[0] for w in self.diff_counts[k]]
            new_words_set.update(set(words))
        print(f'Total {len(new_words_set)} new words')
        for w in new_words_set:
            new_voc.append_token(w)
        print(f'New vocab size: {len(new_voc)}')
        return new_voc
