import os
import pickle

import json
import datasets
import stanza
import torch
from datasets import load_metric
from tqdm import tqdm, trange
import numpy as np

from pysyngen.utils import tree_depth


def copy_baseline(dsn, vocabn='g42B_allds_unk1'):
    os.makedirs(f'Predictions/comparisons', exist_ok=True)
    res = {
        'pred': []
    }
    vocab = torch.load(f"Data/vocab/{vocabn}.pt")
    ds = datasets.load_from_disk(f"Data/ds/{dsn}")['test']

    for s in ds:
        res['pred'].append(vocab.lookup_tokens(s['source'][1:-1]))
    res['vocab'] = vocabn
    res['dsname'] = dsn

    fn_res = f'Predictions/comparisons/copy_{dsn}.res.pkl'
    with open(fn_res, 'wb') as f_res:
        pickle.dump(res, f_res)
    print(fn_res)
    return fn_res


def comparison_convert(fname, comp, ds, skiplist, vocab='g42B_allds_unk1'):
    os.makedirs(f'Predictions/comparisons', exist_ok=True)
    res = {
        'pred': []
    }
    with open(fname, 'r') as f_res:
        lines = f_res.readlines()
        for idx, line in enumerate(lines):
            if idx in skiplist:
                continue
            res['pred'].append(line.rstrip().split(' '))
    res['vocab'] = vocab
    res['dsname'] = ds

    fn_res = f'Predictions/comparisons/{comp}_{ds}.res.pkl'
    with open(fn_res, 'wb') as f_res:
        pickle.dump(res, f_res)
    print(fn_res)
    return fn_res


def ssm_score_greedy(pred, ref):
    pred_syn = sorted(pred, key=lambda x: x[1])
    ref_syn = sorted(ref, key=lambda x: x[1])
    length = len(ref_syn)
    sent_score = 0
    for syn_rel in ref_syn:
        score = 0
        flag = False
        idx = 0
        # find corresponding one in pred_syn
        while idx < len(pred_syn) and not flag:
            if syn_rel[0] != pred_syn[idx][0]:
                idx += 1
                continue
            score += 1 / (abs(syn_rel[1] - pred_syn[idx][1]) + 1)
            flag = True
        if flag:
            pred_syn.pop(idx)
        sent_score += score
    return sent_score / length


def ssm_score_optim(pred, ref):
    pred_syn = sorted(pred, key=lambda x: x[1])
    ref_syn = sorted(ref, key=lambda x: x[1])
    length = len(ref_syn)
    sent_score = 0
    matched_ref = set()
    # first round, search for exact match
    for syn_rel in ref_syn:
        flag = False
        idx = 0
        # find corresponding one in pred_syn
        while idx < len(pred_syn) and not flag:
            if syn_rel == pred_syn[idx]:
                sent_score += 1
                flag = True
                matched_ref.add(syn_rel)
            else:
                idx += 1
                continue
        if flag:
            pred_syn.pop(idx)
    for rel in matched_ref:
        ref_syn.remove(rel)
    # second round, search for rel match with layer diff
    for syn_rel in ref_syn:
        flag = False
        idx = 0
        # find corresponding one in pred_syn
        while idx < len(pred_syn) and not flag:
            if syn_rel[0] == pred_syn[idx][0]:
                sent_score += 1 / (abs(syn_rel[1] - pred_syn[idx][1]) + 1)
                flag = True
            else:
                idx += 1
                continue
        if flag:
            pred_syn.pop(idx)
    return sent_score / length


class PredictionLoader:
    def __init__(self, fn_res):
        with open(fn_res, 'rb') as f_res:
            result_pack = pickle.load(f_res)

        self.predictions = result_pack['pred']

        self.vocab = torch.load(f"Data/vocab/{result_pack['vocab']}.pt")
        self.relvocab = torch.load('Data/vocab/relvocab.pt')
        self.bos_id = self.vocab['<bos>']
        self.eos_id = self.vocab['<eos>']
        self.pad_id = self.vocab['<pad>']

        self.ds = datasets.load_from_disk(f"Data/ds/{result_pack['dsname']}")['test']

        self.src = []
        self.tgt = []
        self.tgt_h = []
        self.tgt_r = []
        self.tgt_dep = []

        self.extract_samples()

        self.res = {
            'pred': self.predictions,
            'tgt': self.tgt,
            'src': self.src,
            'tgt_h': self.tgt_h,
            'tgt_r': self.tgt_r,
            'tgt_dep': self.tgt_dep
        }

    def extract_samples(self):
        for s in self.ds:
            self.src.append(self.wid2word(s['source']))
            self.tgt.append(self.wid2word(s['target']))
            self.tgt_h.append(s['tgt_h'])
            self.tgt_r.append(self.rid2rel(s['tgt_r']))
            self.tgt_dep.append(s['tgt_dep'])

    def wid2word(self, sent):
        if sent[0] == self.bos_id:
            return self.vocab.lookup_tokens(sent[1:])
        else:
            return self.vocab.lookup_tokens(sent)

    def rid2rel(self, sent):
        return self.relvocab.lookup_tokens(sent)


class MultiPredictionEvaluator:
    def __init__(self, bleu_n=4, ibleu_alpha=0.7):
        self.p_bleu_n = bleu_n
        self.ibleu_alpha = ibleu_alpha
        self.met_bleu = load_metric('bleu')
        self.met_bert = load_metric('bertscore')
        self.met_parser = stanza.Pipeline('en', processors='tokenize, pos, lemma, depparse',
                                          tokenize_pretokenized=True)
        self.met_parsed_cache = None

    def evaluate_res(self, fn_res):
        predictions = PredictionLoader(fn_res)
        res_dict = {}
        self.met_parsed_cache = None
        for met in [
            self.bleu_all,
            self.bertscore_ori,
            self.bertscore_ref,
            self.ssm,
            self.dsm
        ]:
            res_dict.update(met(predictions.res))
        self.met_parsed_cache = None
        return res_dict

    def evaluate_multi(self, res_list):
        all_res = {}
        for res_fname in res_list:
            opt_name = res_fname.split('/')[1]
            if opt_name not in all_res:
                all_res[opt_name] = []
            res_name = res_fname.split('/')[-1][:-8]
            print(f'Processing {res_name}')
            eval_res = self.evaluate_res(res_fname)
            print(eval_res)
            all_res[opt_name].append({res_name: eval_res})
        self.save_evaluation(all_res)
        return

    @staticmethod
    def save_evaluation(eval_res_dic):
        for opt, res_lst in eval_res_dic.items():
            if os.path.exists(f'Predictions/{opt}/all_result.json'):
                with open(f'Predictions/{opt}/all_result.json') as f_r:
                    all_res = json.load(f_r)
            else:
                all_res = []
            all_res.extend(res_lst)
            with open(f'Predictions/{opt}/all_result.json', 'w') as f_w:
                json.dump(all_res, f_w)
        return

    def bleu_ref(self, res):
        print(f'Computing BLEU_ref, n={self.p_bleu_n}')
        met_res = self.met_bleu.compute(predictions=res['pred'], references=[[s] for s in res['tgt']],
                                        max_order=self.p_bleu_n)
        return {'bleu_ref': met_res['bleu']}

    def bleu_ori(self, res):
        print(f'Computing BLEU_ori, n={self.p_bleu_n}')
        met_res = self.met_bleu.compute(predictions=res['pred'], references=[[s] for s in res['src']],
                                        max_order=self.p_bleu_n)
        return {'bleu_ori': met_res['bleu']}

    def ibleu_fast(self, res_bleu_ref, res_bleu_ori, alpha=0.5):
        print(f'Computing iBLEU, alpha={self.ibleu_alpha}')
        return {'ibleu': alpha * res_bleu_ref + (1 - alpha) * res_bleu_ori}

    def bleu_all(self, res):
        res_dict = self.bleu_ref(res)
        res_dict.update(self.bleu_ori(res))
        res_dict.update(self.ibleu_fast(res_dict['bleu_ref'], res_dict['bleu_ori']))
        return res_dict

    def bertscore_ref(self, res):
        print(f'Computing BertScore with reference')
        met_res = self.met_bert.compute(predictions=[' '.join(sen) for sen in res['pred']],
                                        references=[' '.join(sen) for sen in res['tgt']], lang='en')
        return {
            'bert_ref_f1': np.average(met_res['f1']),
            'bert_ref_precision': np.average(met_res['precision']),
            'bert_ref_recall': np.average(met_res['recall']),
        }

    def bertscore_ori(self, res):
        print(f'Computing BertScore with original sentence')
        met_res = self.met_bert.compute(predictions=[' '.join(sen) for sen in res['pred']],
                                        references=[' '.join(sen) for sen in res['src']], lang='en')
        return {
            'bert_ori_f1': np.average(met_res['f1']),
            'bert_ori_precision': np.average(met_res['precision']),
            'bert_ori_recall': np.average(met_res['recall']),
        }

    def sentence2synseq(self, sentence):
        parse_res = self.met_parser([sentence])
        head_seq = []
        rel_seq = []
        for w in parse_res.sentences[0].words:
            head_seq.append(w.head)
            rel_seq.append(w.deprel.split(':')[0])
        depth = tree_depth(head_seq)
        return head_seq, rel_seq, depth

    def prep_depparse(self, res):
        pred_h = []
        pred_r = []
        pred_dep = []
        for pred in tqdm(res['pred']):
            try:
                parsed_sent = self.sentence2synseq(pred)
                pred_h.append(parsed_sent[0])
                pred_r.append(parsed_sent[1])
                pred_dep.append(parsed_sent[2])
            except AssertionError:
                print(pred)
                pred_h.append([])
                pred_r.append([])
                pred_dep.append([])
        self.met_parsed_cache = {
            'pred_h': pred_h,
            'pred_r': pred_r,
            'pred_dep': pred_dep
        }

    def dsm(self, res):
        if self.met_parsed_cache is None:
            self.prep_depparse(res)

        pred_h = self.met_parsed_cache['pred_h']
        pred_r = self.met_parsed_cache['pred_r']

        uas = []
        las = []
        uas_cnt = 0
        las_cnt = 0
        sample_length_total = 0

        assert len(pred_h) == len(res['tgt_h'])
        for idx in trange(len(pred_h)):
            sample_length = len(pred_h[idx])
            sample_length_total += sample_length
            try:
                assert len(pred_h[idx]) == len(res['tgt_h'][idx])
                assert len(pred_r[idx]) == len(res['tgt_r'][idx])
            except AssertionError:
                # print(pred_h[idx], res['tgt_h'][idx])
                # print(pred_r[idx], res['tgt_r'][idx])
                # print(res['tgt'][idx], res['pred'][idx])
                uas.append(0.0)
                las.append(0.0)
                continue
            t_uas = 0
            t_las = 0
            for w_idx, (edge_pred, edge_ref) in enumerate(zip(pred_h[idx], res['tgt_h'][idx])):
                if edge_pred == edge_ref:
                    t_uas += 1
                    uas_cnt += 1
                    if pred_r[idx][w_idx] == res['tgt_r'][idx][w_idx]:
                        t_las += 1
                        las_cnt += 1
            uas.append(t_uas / sample_length)
            las.append(t_las / sample_length)

        res_dict = {'dsm_uas_sample': np.average(uas),
                    'dsm_las_sample': np.average(las),
                    'dsm_uas_corpus': uas_cnt / sample_length_total,
                    'dsm_las_corpus': las_cnt / sample_length_total}

        return res_dict

    def ssm(self, res):
        if self.met_parsed_cache is None:
            self.prep_depparse(res)

        pred_syn = []
        for idx in trange(len(self.met_parsed_cache['pred_h'])):
            pred_syn.append(list(zip(self.met_parsed_cache['pred_r'][idx], self.met_parsed_cache['pred_dep'][idx])))

        score_greedy = []
        score_optim = []
        assert len(pred_syn) == len(res['tgt_h'])
        for idx in trange(len(pred_syn)):
            ref_syn = list(zip(res['tgt_r'][idx], res['tgt_dep'][idx]))
            score_greedy.append(ssm_score_greedy(pred_syn[idx], ref_syn))
            score_optim.append(ssm_score_optim(pred_syn[idx], ref_syn))

        return {'ssm_greedy': np.average(score_greedy),
                'ssm_optim': np.average(score_optim)}
