import configargparse as configargparse
import datasets
import numpy as np
import pytorch_lightning as pl
import torch
from torch.optim.lr_scheduler import ReduceLROnPlateau

from pysyngen.networks.syntax import SemSynTextGenerator, SynTextGenerator
from pysyngen.networks.text import VariationalAutoEncoderRNN
from pysyngen.utils import arg_bool

OPTIMIZER_MAP = {
    'adam': torch.optim.Adam,
    'adamw': torch.optim.AdamW,
    'sgd': torch.optim.SGD,
    'rmsprop': torch.optim.RMSprop
}


class TextVAEModel(pl.LightningModule):
    def __init__(self, hparam):
        super(TextVAEModel, self).__init__()
        self.save_hyperparameters(hparam)
        # load vocab
        self.vocab = torch.load(f'Data/vocab/{self.hparams.vocabname}.pt')
        if isinstance(self.hparams.pretrained_we, str):
            self.semvec = self.hparams.pretrained_we
        else:
            self.semvec = None
        # network
        self.n_textvae = VariationalAutoEncoderRNN(self.vocab, self.hparams.rnn, self.hparams.embed_dim,
                                                   self.hparams.hidden_dim,
                                                   self.hparams.num_layers, self.hparams.z_dim,
                                                   self.hparams.dropout, self.semvec, self.hparams.vae)

    def training_step(self, batch, batch_idx):
        loss_rec, loss_kl = self.train_forward(batch)
        if self.hparams.vae:
            kl_step_weight = self.get_kl_weight(self.hparams.kl_anneal, self.global_step, self.hparams.kl_warm)
            loss = self.hparams.w_kl * kl_step_weight * loss_kl \
                   + self.hparams.w_rec * loss_rec
            self.log('loss_kl', loss_kl, on_step=True, on_epoch=False, prog_bar=False, logger=True)
            self.log('kl_step_weight', kl_step_weight, on_step=True, on_epoch=False, prog_bar=False, logger=True)
        else:
            loss = loss_rec
        self.log('train_loss', loss, on_step=True, on_epoch=False, prog_bar=False, logger=True)
        self.log('loss_rec', loss_rec, on_step=True, on_epoch=False, prog_bar=False, logger=True)
        return loss

    def validation_step(self, batch, batch_idx):
        loss_rec, _ = self.train_forward(batch)
        # print(self.training)
        self.log('val_loss', loss_rec, on_step=False, on_epoch=True, prog_bar=True, logger=True)
        return

    def setup_test(self, decode_method=None):
        if decode_method is not None:
            self.hparams.dec_method = decode_method

    def configure_optimizers(self):
        optimizer = OPTIMIZER_MAP[self.hparams.optimizer](filter(lambda p: p.requires_grad, self.parameters()),
                                                          lr=self.hparams.lr,
                                                          weight_decay=self.hparams.weight_decay)
        if self.hparams.lr_sched != 0:
            lr_scheduler = ReduceLROnPlateau(optimizer, patience=self.hparams.lr_sched)
            return {
                'optimizer': optimizer,
                'lr_scheduler': {
                    'scheduler': lr_scheduler,
                    'monitor': 'val_loss',
                    'frequency': 1,
                    'interval': 'epoch'
                }
            }
        else:
            return optimizer

    def get_kl_weight(self, anneal_function, step, warmup=0):
        step = max(0, step - warmup)
        if anneal_function == "fixed":
            return 1.0
        elif anneal_function == 'logistic':
            return float(1 / (1 + np.exp(-self.hparams.kl_k * (step - self.hparams.kl_x0))))
        elif anneal_function == 'negative-sigmoid':
            return float(1 / (1 + np.exp(-self.hparams.kl_k * (self.hparams.kl_x0 - step))))
        elif anneal_function == 'linear':
            return min(1, step / self.hparams.kl_x0)
        elif anneal_function == 'linear_bound':
            return min(0.8, step / self.hparams.kl_x0)

    def train_forward(self, train_inputs):
        # src_seq, src_len, tgt_seq, tgt_len = train_inputs
        loss_rec, loss_kl = self.n_textvae.train_forward(train_inputs['src_seq'], train_inputs['src_len'],
                                                         train_inputs['tgt_seq'], train_inputs['tgt_len'])
        return loss_rec, loss_kl

    def param_name(self, prediction=False):
        return f'{"vae" if self.hparams.vae else "ae"}_{self.hparams.rnn}' \
               f'_w{self.hparams.embed_dim}_h{self.hparams.hidden_dim}_z{self.hparams.z_dim}' \
               f'_l{self.hparams.num_layers}_{self.hparams.kl_anneal}_w{self.hparams.w_kl},{self.hparams.w_rec}'

    def generate(self, strategy, max_length, inputs, inputs_len):
        z, _, _ = self.n_textvae.encode(inputs, inputs_len)
        if strategy == 'greedy':
            results = self.n_textvae.decode_greedy(z, max_length)
        elif strategy.startswith('beam'):
            beam_size = int(strategy[4:])
            results = self.n_textvae.decode_beam(z, beam_size, max_length)
        else:
            results = None
        return results

    @staticmethod
    def setup_parameters():
        def pretrain_we_check(v):
            if v.lower() in ('no', 'false', 'f', 'n', '0'):
                return False
            else:
                return v

        parser = configargparse.get_argument_parser()
        model = parser.add_argument_group('TextVAE Model', 'Model settings')
        # model.add_argument('--vocab', default='Data/vocab/glove6B_unk1.pt', type=str, help='Vocabulary path')
        model.add_argument('--rnn', default='lstm', type=str, help='Type of RNN Cell')
        model.add_argument('--vae', default=True, type=arg_bool, help='VAE or Plain AE')
        model.add_argument('--pretrained_we', default='Data/semvec/glove.6B.300d.txt', type=pretrain_we_check,
                           help='Pretrain word embedding path or False')
        model.add_argument('--embed_dim', default=300, type=int, help='Embedding dimension')
        model.add_argument('--hidden_dim', default=150, type=int, help='Hidden vector dimension')
        model.add_argument('--z_dim', default=50, type=int, help='z dimension')
        model.add_argument('--num_layers', default=2, type=int, help='Number of RNN layers')
        model.add_argument('--dropout', default=0.1, type=float, help='Dropout rate')
        model.add_argument('--kl_anneal', default='sigmoid', type=str, help='KL loss annealing function')
        model.add_argument('--kl_warm', default=0, type=int, help='KL loss annealing warmup steps')
        model.add_argument('--kl_k', default=0.001, type=float, help='KL loss annealing parameter k')
        model.add_argument('--kl_x0', default=2500, type=int, help='KL loss annealing parameter x0')
        model.add_argument('--w_kl', default=10, type=float, help='Weight of KL loss')
        model.add_argument('--w_rec', default=1, type=float, help='Weight of Rec loss')
        optim = parser.add_argument_group('Optimizer settings', 'Settings of the optimizer')
        optim.add_argument('--lr', default=0.004, type=float, help='Learning rate')
        optim.add_argument('--lr_sched', default=1, type=int, help='0 for no scheduling, higher for patience')
        optim.add_argument('--optimizer', default='adam', type=str, help='Optimizer')
        optim.add_argument('--weight_decay', default=4e-4, type=float, help='Weight decay')


class SemSynGeneratorModel(pl.LightningModule):
    def __init__(self, hparam):
        super(SemSynGeneratorModel, self).__init__()
        self.save_hyperparameters(hparam)

        # load vocab
        self.vocab = torch.load(f'Data/vocab/{self.hparams.vocabname}.pt')
        self.relvocab = torch.load('Data/vocab/relvocab.pt')
        if isinstance(self.hparams.pretrained_we, str):
            self.semvec = self.hparams.pretrained_we
        else:
            self.semvec = None
        self.component = torch.load(f'TrainedModel/text/{self.hparams.textvae}/params.pt')
        try:
            tmp = self.hparams.loss_scheme
        except AttributeError:
            self.hparams.loss_scheme = 'plain'
        # network
        self.n_semsyngen = SemSynTextGenerator(self.component,
                                               self.hparams.rnn, self.relvocab, self.vocab, self.hparams.maxlen,
                                               self.hparams.syn_emb_dim, self.hparams.word_emb_dim,
                                               self.hparams.hidden_dim, self.hparams.z_dim,
                                               self.hparams.syn_layers, self.hparams.gen_layers,
                                               self.hparams.dropout, self.hparams.word_dropout,
                                               self.hparams.att_scheme, self.hparams.loss_scheme, self.semvec)

    def training_step(self, batch, batch_idx):
        # {'src_seq': src_seq, 'src_len': src_len,
        #  'tgt_seq': tgt_seq, 'tgt_len': tgt_len,
        #  'tgt_h': tgt_h, 'tgt_r': tgt_r}
        loss_rec, loss_syn = self.train_forward(batch)
        loss = self.hparams.w_rec * loss_rec + self.hparams.w_syn * loss_syn
        self.log('train_loss', loss, on_step=True, on_epoch=False, prog_bar=False, logger=True)
        self.log('loss_rec', loss_rec, on_step=True, on_epoch=False, prog_bar=False, logger=True)
        self.log('loss_syn', loss_syn, on_step=True, on_epoch=False, prog_bar=False, logger=True)
        return loss

    def validation_step(self, batch, batch_idx):
        loss_rec, _ = self.train_forward(batch)
        # print(self.training)
        self.log('val_loss', loss_rec, on_step=False, on_epoch=True, prog_bar=True, logger=True)
        return

    def setup_test(self, decode_method=None):
        if decode_method is not None:
            self.hparams.dec_method = decode_method

    def predict_step(self, batch, batch_idx, **kwargs):
        if self.hparams.dec_method == 'greedy':
            pred, _ = self.n_semsyngen.generate_greedy(batch['src_seq'], batch['src_len'],
                                                       batch['tgt_h'], batch['tgt_r'], batch['tgt_len'])
        elif self.hparams.dec_method.startswith('beam'):
            beam_size = int(self.hparams.dec_method[5:])
            pred, _ = self.n_semsyngen.generate_beam(beam_size, batch['src_seq'], batch['src_len'],
                                                     batch['tgt_h'], batch['tgt_r'], batch['tgt_len'])
        else:
            pred = None
            exit(1)
        return pred

    def train_forward(self, inputs):
        # src_seq, src_len, tgt_seq, tgt_len = train_inputs
        loss_rec, loss_syn = self.n_semsyngen.train_forward(inputs['src_seq'], inputs['src_len'],
                                                            inputs['tgt_h'], inputs['tgt_r'],
                                                            inputs['tgt_seq'], inputs['tgt_len'],
                                                            inputs['tgt_dep'])
        return loss_rec, loss_syn

    def param_name(self, prediction=False):
        pname = f'{self.hparams.rnn}_s{self.hparams.syn_emb_dim}_w{self.hparams.word_emb_dim}' \
                f'_h{self.hparams.hidden_dim}_z{self.hparams.z_dim}' \
                f'_l{self.hparams.syn_layers}.{self.hparams.gen_layers}' \
                f'_w{self.hparams.w_rec},{self.hparams.w_syn}' \
                f'_dp{self.hparams.word_dropout}' \
                f'_{self.hparams.loss_scheme}' \
                f'_{self.hparams.att_scheme if self.hparams.att_scheme is not None else "no"}'
        if prediction:
            pname += f'_{self.hparams.dec_method}'
        return pname

    def configure_optimizers(self):
        optimizer = OPTIMIZER_MAP[self.hparams.optimizer](filter(lambda p: p.requires_grad, self.parameters()),
                                                          lr=self.hparams.lr,
                                                          weight_decay=self.hparams.weight_decay)
        if self.hparams.lr_sched != 0:
            lr_scheduler = ReduceLROnPlateau(optimizer, patience=self.hparams.lr_sched)
            return {
                'optimizer': optimizer,
                'lr_scheduler': {
                    'scheduler': lr_scheduler,
                    'monitor': 'val_loss',
                    'frequency': 1,
                    'interval': 'epoch'
                }
            }
        else:
            return optimizer

    def load_state_dict(self, state_dict, strict: bool = True):
        diff_keys = super().load_state_dict(state_dict, strict=False)
        if "n_semsyngen.n_syngen.len_mask" in diff_keys.missing_keys and \
                "n_semsyngen.n_syngen.length_mask" in diff_keys.unexpected_keys:
            state_dict["n_semsyngen.n_syngen.len_mask"] = state_dict.pop("n_semsyngen.n_syngen.length_mask")
        super().load_state_dict(state_dict, strict=True)

    @staticmethod
    def setup_parameters():
        def str_or_none(v):
            if v.lower() in ('no', 'false', 'f', 'n', '0'):
                return None
            else:
                return v

        parser = configargparse.get_argument_parser()
        model = parser.add_argument_group('Semantic Syntac Text Generator', 'Model settings')
        # model.add_argument('--vocab', default='Data/vocab/glove6B_unk1.pt', type=str, help='Vocabulary path')
        model.add_argument('--rnn', default='lstm', type=str, help='Type of RNN Cell')
        model.add_argument('--pretrained_we', default='Data/semvec/glove.6B.300d.txt', type=str_or_none,
                           help='Pretrain word embedding path or False')
        model.add_argument('--textvae', default='pt_swiki', type=str, help='Text Encoder params')
        model.add_argument('--syn_emb_dim', default=300, type=int, help='Syntax embedding dimension')
        model.add_argument('--word_emb_dim', default=300, type=int, help='Word embedding dimension')
        model.add_argument('--hidden_dim', default=150, type=int, help='Hidden vector dimension')
        model.add_argument('--z_dim', default=50, type=int, help='z dimension')
        model.add_argument('--syn_layers', default=2, type=int, help='Number of RNN layers in syntax encoder')
        model.add_argument('--gen_layers', default=2, type=int, help='Number of RNN layers in generator')
        model.add_argument('--word_dropout', default=0.4, type=float, help='Dropout rate for source words')
        model.add_argument('--dropout', default=0.1, type=float, help='Dropout rate')
        model.add_argument('--att_scheme', default='no', type=str_or_none, help='Attention scheme')
        model.add_argument('--loss_scheme', default='plain', type=str, help='Syntax loss scheme')
        model.add_argument('--w_syn', default=1, type=float, help='Weight of Syntax loss')
        model.add_argument('--w_rec', default=1, type=float, help='Weight of Rec loss')
        optim = parser.add_argument_group('Optimizer settings', 'Settings of the optimizer')
        optim.add_argument('--lr', default=0.004, type=float, help='Learning rate')
        optim.add_argument('--lr_sched', default=1, type=int, help='0 for no scheduling, higher for patience')
        optim.add_argument('--optimizer', default='adam', type=str, help='Optimizer')
        optim.add_argument('--weight_decay', default=4e-4, type=float, help='Weight decay')
        test = parser.add_argument_group('Prediction settings', 'Settings of the testing')
        test.add_argument('--dec_method', default='greedy', type=str, help='Decoding method. "greedy" or "beam.size"')
