import math

import torch
from torch import nn
from torch.nn import functional as F

from pysyngen.utils import read_vectors


class PlainLSTM(nn.Module):
    def __init__(self, input_dim, hidden_dim, num_layers, bidirectional, dropout):
        super(PlainLSTM, self).__init__()
        self.lstm = nn.LSTM(input_size=input_dim,
                            hidden_size=hidden_dim,
                            num_layers=num_layers,
                            batch_first=True,
                            bidirectional=bidirectional,
                            dropout=dropout)
        self.bidir = bidirectional
        self.num_layers = num_layers
        self.hidden_size = hidden_dim

    def forward(self, x, x_len=None, ctx=None):
        self.lstm.flatten_parameters()
        if x_len is not None:
            x_len, x_idx = x_len.sort(descending=True)
            x = x[x_idx]
            x = nn.utils.rnn.pack_padded_sequence(x, x_len.to('cpu'), batch_first=True)
            if ctx is not None:
                ctx = (ctx[0][:, x_idx, :], ctx[1][:, x_idx, :])
        o, (h, c) = self.lstm(x) if ctx is None else self.lstm(x, ctx)
        if x_len is not None:
            o, _ = nn.utils.rnn.pad_packed_sequence(o, batch_first=True)
            _, x_idx = x_idx.sort(descending=False)
            o = o[x_idx]
            h = h[:, x_idx, :]
            c = c[:, x_idx, :]
        return o, (h, c)

    @staticmethod
    def indexing_ctx(ctx, indexes):
        h = ctx[0][:, indexes, :]
        c = ctx[1][:, indexes, :]
        return h, c

    @staticmethod
    def setting_ctx(ctx, index, new_ctx):
        ctx[0][:, index, :] = new_ctx[0]
        ctx[1][:, index, :] = new_ctx[1]
        return ctx


class PlainGRU(nn.Module):
    def __init__(self, input_dim, hidden_dim, num_layers, bidirectional, dropout):
        super(PlainGRU, self).__init__()
        self.gru = nn.GRU(input_size=input_dim,
                          hidden_size=hidden_dim,
                          num_layers=num_layers,
                          batch_first=True,
                          bidirectional=bidirectional,
                          dropout=dropout)
        self.bidir = bidirectional
        self.num_layers = num_layers
        self.hidden_size = hidden_dim

    def forward(self, x, x_len=None, ctx=None):
        self.gru.flatten_parameters()
        if x_len is not None:
            x_len, x_idx = x_len.sort(descending=True)
            x = x[x_idx]
            x = nn.utils.rnn.pack_padded_sequence(x, x_len.to('cpu'), batch_first=True)
            if ctx is not None:
                ctx = ctx[:, x_idx, :]
        o, h = self.gru(x) if ctx is None else self.gru(x, ctx)
        if x_len is not None:
            o, _ = nn.utils.rnn.pad_packed_sequence(o, batch_first=True)
            _, x_idx = x_idx.sort(descending=False)
            o = o[x_idx]
            h = h[:, x_idx, :]
        return o, h

    @staticmethod
    def indexing_ctx(ctx, indexes):
        return ctx[:, indexes, :]

    @staticmethod
    def setting_ctx(ctx, index, new_ctx):
        ctx[:, index, :] = new_ctx
        return ctx


class PartialTrainableEmbedding(nn.Module):
    def __init__(self, semvec, word_dim, splen):
        super(PartialTrainableEmbedding, self).__init__()
        self.embedding_dim = word_dim
        if isinstance(semvec, int):
            # a reserved testing mode, which is completely unintended since n_word should be un-trainable
            # semvec should be word count + 1 for padding though, to make sense
            self.n_word = nn.Embedding(semvec, word_dim, padding_idx=0)
        else:
            tmp_semvec, self.d_iw, self.d_wi = read_vectors(semvec, word_dim, True)
            # <zero> + words
            self.n_word = nn.Embedding.from_pretrained(tmp_semvec, padding_idx=0)
        # sp
        self.n_spec = nn.Embedding(splen + 1, word_dim, padding_idx=-1)
        with torch.no_grad():
            self.n_spec.weight[0].fill_(0)

        self.num_sp = splen
        self.num_word = self.n_word.num_embeddings - 1
        self.num_embeddings = self.n_spec.num_embeddings + self.n_word.num_embeddings - 2

    def forward(self, inputs):
        # inputs: sp + words
        # in n_word, need to sub sp-1 to get words for words and 0 for sp.
        # in n_sp, need to max as sp (counting as sp+1) to get sp for sp and padding_idx(-1) for word.
        if isinstance(inputs, torch.LongTensor) or (
                torch.cuda.is_available() and isinstance(inputs, torch.cuda.LongTensor)
        ):
            return self.n_word(torch.clamp(inputs - self.num_sp + 1, min=0)) + \
                   self.n_spec(torch.clamp(inputs, max=self.num_sp))
        else:
            # it's easier for mm since we deal with actual positions, just skip the paddings
            return torch.matmul(inputs[..., :self.num_sp], self.n_spec.weight[:self.num_sp, :]) + \
                   torch.matmul(inputs[..., self.num_sp:], self.n_word.weight[1:, :])


class QtoKVAttention(nn.Module):
    def __init__(self, q_input_dim, kv_input_dim, hidden_dim, out_dim):
        super(QtoKVAttention, self).__init__()
        self.n_q_linear = nn.Linear(q_input_dim, hidden_dim)
        self.n_k_linear = nn.Linear(kv_input_dim, hidden_dim)
        self.n_v_linear = nn.Linear(kv_input_dim, out_dim)
        self.scale = math.sqrt(hidden_dim)
        self.out_dim = out_dim

    def forward(self, q, kv, kv_len):
        # Nq == Nkv
        # B, Nq, Eq->Eh
        q = self.n_q_linear(q)
        # B, Nkv, Ekv->Eh
        k = self.n_k_linear(kv)
        # B, Nkv, Ekv->Eo
        v = self.n_v_linear(kv)
        # make attn mask with Nq and kv_len -> (B, Nq, Nkv)
        bs = k.size(0)
        maxlen = k.size(1)
        # attn_mask_b = torch.arange(maxlen).expand(bs, maxlen).to(q.device)
        # attn_mask_b = attn_mask_b >= kv_len.unsqueeze(1)
        # attn_mask_b = attn_mask_b.unsqueeze(1).expand(-1, q.size(1), -1)
        attn_mask_b = (torch.arange(maxlen).expand(bs, maxlen).to(q.device) >= kv_len.unsqueeze(1)) \
            .unsqueeze(1).expand(-1, q.size(1), -1)
        attn_mask = torch.zeros_like(attn_mask_b, dtype=torch.float).to(q.device)
        # mask: B, Nkv, Nkv
        attn_mask.masked_fill_(attn_mask_b, -torch.inf)

        # (B, Nq, Eh) x (B, Eh, Nkv) -> (B, Nq, Nkv)
        attn = torch.bmm(q, k.transpose(-2, -1))
        attn = torch.div(attn, self.scale)
        attn += attn_mask
        attn = F.softmax(attn, dim=-1)
        # (B, Nkv, Nkv) x (B, Nkv, Eo) -> (B, Nkv, Eo)
        output = torch.bmm(attn, v)
        return output, attn


class DotAttention(nn.Module):
    def __init__(self, q_input_dim, kv_input_dim, hidden_dim, out_dim):
        super(DotAttention, self).__init__()
        self.n_q_linear = nn.Linear(q_input_dim, kv_input_dim)
        self.out_dim = kv_input_dim

    def forward(self, q, kv, kv_len):
        # Nq == Nkv
        # B, Nkv, Eq->Eh
        q = self.n_q_linear(q)
        # make attn mask with kv_len
        bs = kv.size(0)
        maxlen = kv.size(1)
        attn_mask_b = (torch.arange(maxlen).expand(bs, maxlen).to(q.device) >= kv_len.unsqueeze(1)) \
            .unsqueeze(1).expand(-1, maxlen, -1)
        attn_mask = torch.zeros_like(attn_mask_b, dtype=torch.float).to(q.device)
        # mask: B, Nkv, Nkv
        attn_mask.masked_fill_(attn_mask_b, -torch.inf)

        # (B, Nkv, Eh) x (B, Eh, Nkv) -> (B, Nkv, Nkv)
        attn = torch.bmm(q, kv.transpose(-2, -1))
        attn += attn_mask
        attn = F.softmax(attn, dim=-1)
        # (B, Nkv, Nkv) x (B, Nkv, Eo) -> (B, Nkv, Eo)
        output = torch.bmm(attn, kv)
        return output, attn


def gumbel_softmax(logits, tau: float = 1, hard: bool = False, dim: int = -1):
    gumbels = (
        -torch.empty_like(logits, memory_format=torch.legacy_contiguous_format).exponential_().log()
    )  # ~Gumbel(0,1)
    gumbels = (logits + gumbels) / tau  # ~Gumbel(logits,tau)
    y_soft = gumbels.softmax(dim)

    if hard:
        # Straight through.
        index = y_soft.max(dim, keepdim=True)[1]
        y_hard = torch.zeros_like(logits, memory_format=torch.legacy_contiguous_format).scatter_(dim, index, 1.0)
        ret = y_hard - y_soft.detach() + y_soft
    else:
        # Reparametrization trick.
        ret = y_soft
    return ret
