import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.nn import functional as F, Parameter

from pysyngen.networks.common import PartialTrainableEmbedding, PlainLSTM, PlainGRU, QtoKVAttention, gumbel_softmax, \
    DotAttention
from pysyngen.networks.synvec import MiniSynVecModel
from pysyngen.networks.text import VariationalEncoderRNN


class SynTextGenerator(nn.Module):
    def __init__(self, rnn, relvocab, wordvocab, maxlen,
                 syn_embed_dim, word_embed_dim,
                 hidden_dim, o_z_dim, z_dim,
                 syn_layers, gen_layers,
                 dropout, att_scheme=None, loss_scheme='plain',
                 semvec=None, synvec=None):
        super(SynTextGenerator, self).__init__()
        # record parameters
        self.p_rnn = rnn
        self.p_att_scheme = att_scheme
        self.p_loss_scheme = loss_scheme

        self.p_relsz = len(relvocab)
        self.p_vocsz = len(wordvocab)
        self.p_maxlen = maxlen

        self.p_syn_dim = syn_embed_dim
        self.p_word_dim = word_embed_dim
        self.p_hidden_dim = hidden_dim
        self.p_o_z_dim = o_z_dim
        self.p_z_dim = z_dim

        self.p_syn_layers = syn_layers
        self.p_gen_layers = gen_layers

        self.p_dropout = dropout

        # vocabs
        self.vocab_word = wordvocab
        self.vocab_sz = len(self.vocab_word)
        self.bos_id = self.vocab_word['<bos>']
        self.eos_id = self.vocab_word['<eos>']
        self.pad_id = self.vocab_word['<pad>']
        current_unk_id = self.vocab_word['<unk>']
        for i in range(1, len(self.vocab_word) - 3):
            tmp_unk_id = self.vocab_word[f'<unk{i}>']
            if tmp_unk_id <= current_unk_id:
                break
            else:
                current_unk_id = tmp_unk_id
        self.splen = current_unk_id + 1

        self.vocab_rel = relvocab  # ?

        # Syntax tree encoder

        self.n_syn_rel_emb = nn.Embedding(len(relvocab), syn_embed_dim)
        self.n_syn_idx_emb = nn.Embedding(self.p_maxlen + 2, syn_embed_dim)
        # positional embedding?
        if self.p_rnn == 'lstm':
            self.n_syn_enc = PlainLSTM(self.p_syn_dim * 2, self.p_syn_dim, self.p_syn_layers,
                                       True, self.p_dropout)
        elif self.p_rnn == 'gru':
            self.n_syn_enc = PlainGRU(self.p_syn_dim * 2, self.p_syn_dim, self.p_syn_layers,
                                      True, self.p_dropout)

        # Text generator

        if semvec is None:
            # we use a fully trainable embedding layer as default
            self.n_emb = nn.Embedding(self.vocab_sz, self.p_word_dim)
        else:
            self.n_emb = PartialTrainableEmbedding(semvec, self.p_word_dim, self.splen)
        # project z
        if self.p_o_z_dim != self.p_z_dim:
            self.n_z_proj = nn.Linear(self.p_o_z_dim, self.p_z_dim)
        else:
            self.n_z_proj = None
        # vocab_mask
        vocab_mask = torch.zeros(self.vocab_sz)
        vocab_mask[:3] = -torch.inf
        self.register_buffer('vocab_mask', vocab_mask)
        # plain version decoder input dim
        input_dim = self.p_word_dim + self.p_z_dim + self.p_syn_dim * 2
        if self.p_att_scheme is not None and self.p_att_scheme.startswith('dot'):
            self.p_att_scheme = self.p_att_scheme[3:]
            attn = DotAttention
        else:
            attn = QtoKVAttention
        # attention
        if self.p_att_scheme == 'wsz':
            # Q: word+H_syn_i+z, K: H_syn, V: H_syn
            self.n_attn = attn(input_dim, self.p_syn_dim * 2, self.p_syn_dim, self.p_syn_dim)
            input_dim += self.n_attn.out_dim
        elif self.p_att_scheme == 'wz':
            # Q: word+z, K: H_syn, V: H_syn
            self.n_attn = attn(input_dim - self.p_syn_dim * 2, self.p_syn_dim * 2, self.p_syn_dim, self.p_syn_dim)
            input_dim += self.n_attn.out_dim
        elif self.p_att_scheme == 'wz.2':
            # Q: word+z, K: H_syn, V: H_syn, shrink hidden
            self.n_attn = attn(input_dim - self.p_syn_dim * 2, self.p_syn_dim * 2, self.p_syn_dim // 2, self.p_syn_dim)
            input_dim += self.n_attn.out_dim
        else:
            self.n_attn = None
        if self.p_rnn == 'lstm':
            self.n_text_dec = PlainLSTM(input_dim, self.p_hidden_dim, self.p_gen_layers, False, self.p_dropout)
        elif self.p_rnn == 'gru':
            self.n_text_dec = PlainGRU(input_dim, self.p_hidden_dim, self.p_gen_layers, False, self.p_dropout)

        self.n_text_dec_out = nn.Linear(self.p_hidden_dim, self.p_vocsz)

        # Loss function
        self.tau = 1
        self.len_mask = Parameter(torch.arange(0, maxlen), requires_grad=False)
        self.synvec_model = synvec if synvec is not None else MiniSynVecModel.get_default()

        # temp variables to prevent re-creation
        self.len_1 = torch.LongTensor([1] * 32)

    def train_forward(self, z, tgt_h, tgt_r, tgt_seq, tgt_len, tgt_dep):
        batch_size = z.size(0)
        H_syn = self.encode_syn(tgt_h, tgt_r, tgt_len - 1)
        tgt_pred = self.decode_train(z, H_syn, tgt_seq[:, :-1], tgt_len - 1)
        # reconstruction loss
        tgt_seq_rec = tgt_seq[:, 1:].contiguous().view(-1)
        tgt_pred_logits = F.log_softmax(tgt_pred, dim=-1)
        loss_rec = torch.gather(tgt_pred_logits.view(-1, tgt_pred_logits.size(2)),
                                1,
                                tgt_seq_rec.unsqueeze(1)).squeeze(1)
        loss_rec = - torch.sum(loss_rec * (1.0 - torch.eq(tgt_seq_rec, self.pad_id).float())) \
                   / (tgt_len.sum().item() - batch_size)
        # syntax loss
        loss_syn = self.loss_syn(tgt_pred, tgt_seq, tgt_len, tgt_h, tgt_r, tgt_dep) / batch_size
        return loss_rec, loss_syn

    def encode_syn(self, tgt_h, tgt_r, syn_len):
        syn_h, _ = self.n_syn_enc(torch.cat([self.n_syn_idx_emb(tgt_h),
                                             self.n_syn_rel_emb(tgt_r)], dim=-1),
                                  syn_len)
        return syn_h

    def loss_syn(self, tgt_pred, tgt_seq, tgt_len, tgt_h, tgt_r, tgt_dep):
        # convert tgt_pred to embeddings
        tgt_pred = tgt_pred + self.vocab_mask.expand_as(tgt_pred)
        tgt_pred = self.n_emb(gumbel_softmax(tgt_pred, self.tau))[..., :self.synvec_model.word_dim]
        len_mask = self.len_mask[:tgt_h.size(1)].unsqueeze(0).expand(tgt_len.size(0), -1) < (tgt_len - 1).unsqueeze(1)
        # prepare loss weights
        if self.p_loss_scheme != 'plain':
            max_dep = tgt_dep.max(dim=1)[0].unsqueeze(1)
            w_mat = (max_dep + 1 - tgt_dep).masked_fill(len_mask.logical_not(), 0).float()
        # get normal coords
        normal_coord = tgt_h.nonzero(as_tuple=True)
        # get normal heads and their embeddings since not everybody can be head
        # normal_head = self.n_emb(torch.gather(tgt_seq, 1, syn_h_seq)[normal_coord])
        #               [:self.loss_synvec.word_dim].detach()
        # normal_rel = syn_rel_seq[normal_coord]
        # normal_pred = tgt_pred[normal_coord]
        # score normal words
        normal_loss = self.synvec_model(self.n_emb(torch.gather(tgt_seq, 1, tgt_h)[normal_coord]
                                                   )[..., :self.synvec_model.word_dim].detach(),
                                        tgt_pred[normal_coord],
                                        tgt_r[normal_coord]).squeeze()
        # get root coords
        root_coord = ((tgt_h == 0) * len_mask).nonzero(as_tuple=True)
        # get root pred
        # root_pred = tgt_pred[root_coord]
        # score root words
        root_loss = self.synvec_model.root_score(tgt_pred[root_coord]).squeeze()

        if self.p_loss_scheme == 'normal':
            w_mat = w_mat / max_dep
        elif self.p_loss_scheme == 'exp':
            # w_mat = F.softmax(w_mat.masked_fill(len_mask.logical_not(), -torch.inf), dim=-1)
            w_mat = w_mat.exp() / (max_dep + 1).exp()
        if self.p_loss_scheme != 'plain':
            normal_loss = normal_loss * w_mat[normal_coord]
            root_loss = root_loss * w_mat[root_coord]

        loss = - torch.cat([root_loss, normal_loss], dim=0).mean()
        # combine and return
        return loss

    def decode_train(self, z, H_syn, tgt_seq, tgt_len):
        if self.n_z_proj is not None:
            z = self.n_z_proj(z)
        assert H_syn.size(1) == tgt_len.max().item()
        x = self.n_emb(tgt_seq)
        x = torch.cat([
            x,
            z.unsqueeze(1).expand(-1, x.size(1), -1)
        ], dim=-1)

        att_seq = None
        if self.p_att_scheme == 'wz' or self.p_att_scheme == 'wz.2':
            att_seq, _ = self.n_attn(x, H_syn, tgt_len)

        x = torch.cat([x, H_syn], dim=-1)

        if self.p_att_scheme == 'wsz':
            att_seq, _ = self.n_attn(x, H_syn, tgt_len)
        if att_seq is not None:
            x = torch.cat([x, att_seq], dim=-1)

        out_seq, _ = self.n_text_dec(x, tgt_len)
        out_logits = self.n_text_dec_out(out_seq)
        return out_logits

    def decode_step(self, z, H_syn, i, tgt_step, tgt_len, ctx=None):
        if self.n_z_proj is not None:
            z = self.n_z_proj(z)
        assert H_syn.size(1) == tgt_len.max().item()
        batch_size = z.size(0)
        if self.len_1.size(0) < batch_size:
            self.len_1 = torch.LongTensor([1] * batch_size)
        len_1 = self.len_1[:batch_size]
        x = self.n_emb(tgt_step)
        # print(x.shape, H_syn[:, i].shape, z.shape)
        x = torch.cat([
            x,
            z.unsqueeze(1)
        ], dim=-1)

        att_seq = None
        if self.p_att_scheme == 'wz' or self.p_att_scheme == 'wz.2':
            att_seq, _ = self.n_attn(x, H_syn, tgt_len)

        x = torch.cat([x, H_syn[:, i].unsqueeze(1)], dim=-1)

        if self.p_att_scheme == 'wsz':
            att_seq, _ = self.n_attn(x, H_syn, tgt_len)
        if att_seq is not None:
            x = torch.cat([x, att_seq], dim=-1)

        out_seq, out_ctx = self.n_text_dec(x, len_1, ctx=ctx)
        out_logits = F.softmax(self.n_text_dec_out(out_seq), dim=-1)
        return out_logits, out_ctx

    def decode_greedy(self, z, tgt_h, tgt_r, tgt_len):
        assert tgt_h.size(1) == tgt_len.max().item()
        batch_size = z.size(0)
        translations = [[] for i in range(batch_size)]
        pending = set(range(batch_size))
        prev_words = batch_size * [self.bos_id]
        out_ctx = None
        H_syn = self.encode_syn(tgt_h, tgt_r, tgt_len)
        current_i = 0
        while len(pending) > 0:
            input_seq = Variable(torch.LongTensor([prev_words]).to(z.device)).transpose(0, 1)
            out_logits, out_ctx = self.decode_step(z, H_syn, current_i, input_seq, tgt_len, out_ctx)
            prev_words = out_logits.max(dim=2)[1].squeeze().data.detach().tolist()
            for i in pending.copy():
                # if prev_words[i] == self.eos_id:
                #     pending.discard(i)
                # else:
                translations[i].append(prev_words[i])
                if len(translations[i]) >= tgt_len[i].item():
                    pending.discard(i)
            current_i += 1
        translations = self.id2word(translations)
        return translations

    def decode_beam(self, z, tgt_h, tgt_r, tgt_len, beam_size):
        assert tgt_h.size(1) == tgt_len.max().item()
        batch_size = z.size(0)
        # [beam1 (s0, s1, ..., sbs), beam2 (s0, s1, ..., sbs), ... beamN]
        z = z.repeat(beam_size, 1)
        len_1 = torch.LongTensor([1] * batch_size * beam_size)
        pending = set(range(batch_size))
        prev_words = batch_size * beam_size * [self.bos_id]
        out_ctx = None
        H_syn = self.encode_syn(tgt_h, tgt_r, tgt_len).repeat(beam_size, 1, 1)
        tgt_len = tgt_len.repeat(beam_size)
        current_i = 0

        translations = [[] for i in range(batch_size)]
        translation_scores = batch_size * [-float('inf')]
        # arrangements: top for every sentence in batch, then no2, then no3....
        hypotheses = batch_size * [(0.0, [])] + (beam_size - 1) * batch_size * [
            (-float('inf'), [])]  # (score, translation)

        while len(pending) > 0:
            # run through decoder
            input_seq = Variable(torch.LongTensor([prev_words]).to(z.device)).transpose(0, 1)
            out_logits, out_ctx = self.decode_step(z, H_syn, current_i, input_seq, tgt_len, out_ctx)
            # top 1 word
            # [beam_size*batch_size]
            prev_words = out_logits.max(dim=2)[1].squeeze().data.detach().tolist()
            # top beam words
            # [beam_size*batch_size, beam+1]
            word_scores, words = out_logits.topk(k=beam_size + 1, dim=2, sorted=False)
            word_scores = word_scores.squeeze(1).data.detach().tolist()
            # print(word_scores)
            words = words.squeeze(1).data.detach().tolist()

            for sentence_idx in pending.copy():
                # deal with each sentence
                candidates = []  # (score, index, word)
                for rank in range(beam_size):
                    beam_idx = sentence_idx + rank * batch_size
                    # for each beam we are looking for beam_size+1 candidates for next words
                    # calculate their score and store them in candidates[]
                    for i in range(beam_size + 1):
                        word = words[beam_idx][i]
                        # existing score + current word score
                        score = hypotheses[beam_idx][0] + word_scores[beam_idx][i]
                        # if word != self.eos_id:
                        candidates.append((score, beam_idx, word))
                        # elif score > translation_scores[sentence_idx]:
                        # eos ? shouldn't appear here but listed
                        # translations[sentence_idx] = hypotheses[beam_idx][1] + [word]
                        # translation_scores[sentence_idx] = score
                # beam_size * (beam_size + 1) candidates are listed, looking for top (beam_size) candidates
                best = []  # score, word, translation, hidden, output
                for score, beam_index, word in sorted(candidates, reverse=True)[:beam_size]:
                    translation = hypotheses[beam_index][1] + [word]
                    # best.append(
                    #     (score, word, translation, hidden[:, current_index, :].data, output[current_index].data))
                    best.append(
                        (score, word, translation, self.n_text_dec.indexing_ctx(out_ctx, beam_index)))
                for rank, (score, word, translation, h) in enumerate(best):
                    new_beam_index = sentence_idx + rank * batch_size
                    hypotheses[new_beam_index] = (score, translation)
                    prev_words[new_beam_index] = word
                    out_ctx = self.n_text_dec.setting_ctx(out_ctx, new_beam_index, h)
                # my top option is longer than the length or my top option is worse than existing best translation
                if len(hypotheses[sentence_idx][1]) >= tgt_len[sentence_idx].item():
                    #    or translation_scores[sentence_idx] > hypotheses[sentence_idx][0]:
                    pending.discard(sentence_idx)
                    if len(translations[sentence_idx]) == 0:
                        translations[sentence_idx] = hypotheses[sentence_idx][1]
                        translation_scores[sentence_idx] = hypotheses[sentence_idx][0]
            current_i += 1
        # res = [' '.join(s).replace('@@ ', '') for s in res]
        translations = self.id2word(translations)
        return translations, translation_scores

    def id2word(self, sentences):
        words = []
        for sent in sentences:
            len_i = len(sent)
            while sent[len_i - 1] == self.pad_id:
                len_i -= 1
            if sent[0] == self.bos_id:
                words.append(self.vocab_word.lookup_tokens(sent[1:len_i]))
            else:
                words.append(self.vocab_word.lookup_tokens(sent[:len_i]))
        return words


class SemSynTextGenerator(nn.Module):
    def __init__(self, components,
                 rnn, relvocab, vocab, maxlen,
                 syn_emb_dim, word_emb_dim, hidden_dim, z_dim,
                 syn_layers, gen_layers, dropout, word_dropout,
                 att_scheme, loss_scheme, semvec):
        super(SemSynTextGenerator, self).__init__()
        self.n_semenc = VariationalEncoderRNN(**components['enc_param'])
        self.n_semenc.load_state_dict(components['enc_state_dict'])
        self.n_semenc.freeze_me()
        self.p_ori_z_dim = self.n_semenc.z_dim
        self.p_word_dropout = word_dropout

        self.n_syngen = SynTextGenerator(rnn, relvocab, vocab, maxlen,
                                         syn_emb_dim, word_emb_dim,
                                         hidden_dim, self.p_ori_z_dim, z_dim,
                                         syn_layers, gen_layers,
                                         dropout, att_scheme, loss_scheme, semvec)
        if components['sp_emb'] is not None:
            self.n_syngen.n_emb.n_spec.load_state_dict(components['sp_emb'])

    def train_forward(self, src_seq, src_len, tgt_idx_seq, tgt_rel_seq, tgt_seq, tgt_len, tgt_dep):
        # word dropout
        # god bless i wont mess up with bos and eos?
        if self.p_word_dropout > 0:
            bs = src_seq.size(0)
            maxlen = src_seq.size(1)
            drop_mask = torch.cat([torch.zeros(bs).bool().unsqueeze(1).to(src_seq.device),
                                   torch.arange(maxlen - 1).expand(bs, maxlen - 1).to(src_seq.device) < (
                                           src_len.unsqueeze(1) - 2)], dim=1)
            drop_mask = torch.bernoulli(self.p_word_dropout * torch.ones(bs, maxlen)).bool() \
                .to(src_seq.device).logical_and(drop_mask)
            src_seq = src_seq.masked_fill(drop_mask, 0)

        src_seq = self.n_syngen.n_emb(src_seq)
        z = self.n_semenc(src_seq, src_len, True)
        loss_rec, loss_syn = self.n_syngen.train_forward(z, tgt_idx_seq, tgt_rel_seq,
                                                         tgt_seq, tgt_len, tgt_dep)
        return loss_rec, loss_syn

    def generate_greedy(self, src_seq, src_len, tgt_h, tgt_r, tgt_len):
        src_seq = self.n_syngen.n_emb(src_seq)
        z = self.n_semenc(src_seq, src_len, True)
        res = self.n_syngen.decode_greedy(z, tgt_h, tgt_r, tgt_len - 1)
        return res, 0

    def generate_beam(self, beam_size, src_seq, src_len, tgt_h, tgt_r, tgt_len):
        src_seq = self.n_syngen.n_emb(src_seq)
        z = self.n_semenc(src_seq, src_len, True)
        res, res_score = self.n_syngen.decode_beam(z, tgt_h, tgt_r, tgt_len - 1, beam_size)
        return res, res_score
