import functools

import torch
import torch.nn as nn
from torch.nn import Parameter

from pysyngen.networks.common import PartialTrainableEmbedding


class MiniSynVecModel(nn.Module):
    def __init__(self, rel_tot=36, rel_dim=50, word_dim=150, biased=True):
        super(MiniSynVecModel, self).__init__()

        self.rel_tot = rel_tot
        self.rel_dim = rel_dim
        self.word_dim = word_dim
        self.biased = biased

        self.n_root_v = Parameter(torch.Tensor(self.word_dim), False)
        self.rel_root = Parameter(torch.LongTensor([0]), False)

        self.n_head_trans = RelNonLinearTrans(self.rel_tot, self.word_dim,
                                              self.rel_dim, self.biased)
        self.n_tail_trans = RelNonLinearTrans(self.rel_tot, self.word_dim,
                                              self.rel_dim, self.biased)

        self.n_energy = BilinearEnergy(self.rel_dim)

        self.n_activator = nn.LogSigmoid()

    def load_from_whole_state_dict(self, prev_state_dict):
        self.n_root_v.data = prev_state_dict['n_word_e.n_spec.weight'][0]
        my_state_dict = self.state_dict()
        prev_state_dict = {k: v for k, v in prev_state_dict.items() if k in my_state_dict}
        my_state_dict.update(prev_state_dict)
        self.load_state_dict(my_state_dict)

    def forward(self, emb_head, emb_tail, relations):
        ue = self.n_head_trans(emb_head, relations)
        ve = self.n_tail_trans(emb_tail, relations)
        score = self.n_activator(self.n_energy(ue, ve))
        # print(score.shape)
        return score

    def root_score(self, emb_tail):
        root_rel = self.rel_root.expand(emb_tail.size(0), -1)
        emb_root = self.n_root_v.expand(emb_tail.size(0), -1)
        return self(emb_root, emb_tail, root_rel)

    def loss_1(self, pos_sample, neg_samples):
        # pos_sample: (h_emb, t_emb, r)
        # neg_sample: [(h_emb, t_emb, r)]
        pos_score = self(*pos_sample).sum()
        neg_score = 0
        for neg_samp in neg_samples:
            neg_score += self.n_activator(-self.n_energy(self.n_head_trans(neg_samp[0], neg_samp[2]),
                                                         self.n_tail_trans(neg_samp[1], neg_samp[2])))
        return -(pos_score + neg_score)

    def loss_2(self, pos_sample, neg_r):
        # pos_sample: (h_emb, t_emb, r)
        # neg_r: [r]
        ell_2 = 0
        uhr = self.n_head_trans(pos_sample[0], pos_sample[2])
        vtr = self.n_tail_trans(pos_sample[1], pos_sample[2])
        for rt in neg_r:
            uhrt = self.n_head_trans(pos_sample[0], rt)
            vtrt = self.n_tail_trans(pos_sample[1], rt)
            ell_2 += -(self.n_activator(-self.n_energy(uhr, vtrt)) + self.n_activator(-self.n_energy(uhrt, vtr))).sum()
        return ell_2

    def freeze_me(self):
        for p in self.parameters():
            p.requires_grad = False

    @staticmethod
    def get_default():
        mmodel = MiniSynVecModel()
        mmodel.load_state_dict(torch.load('Data/glove42B_minisynvec.pt'))
        mmodel.freeze_me()
        return mmodel


class RelNonLinearTrans(nn.Module):
    def __init__(self, rel_tot, word_dim=100, rel_dim=100, bias=False):
        super().__init__()
        self.biased = bias
        self.word_dim = word_dim
        self.rel_dim = rel_dim
        self.weight1 = nn.Embedding(rel_tot, word_dim * rel_dim)
        self.weight2 = nn.Embedding(rel_tot, rel_dim * rel_dim)
        if self.biased:
            self.bias1 = nn.Embedding(rel_tot, rel_dim)
            self.bias2 = nn.Embedding(rel_tot, rel_dim)
        # self.activator = nn.LeakyReLU(0.05)
        self.activator = nn.Softplus()

    def forward(self, e, r):
        linear1_w = self.weight1(r).view(-1, self.word_dim, self.rel_dim)
        linear2_w = self.weight2(r).view(-1, self.rel_dim, self.rel_dim)
        e = e.view(-1, 1, self.word_dim)
        if self.biased:
            linear1_b = self.bias1(r).view(-1, 1, self.rel_dim)
            linear2_b = self.bias2(r).view(-1, 1, self.rel_dim)
            e = torch.bmm(self.activator(torch.bmm(e, linear1_w) + linear1_b), linear2_w) + linear2_b
        else:
            e = torch.bmm(self.activator(torch.bmm(e, linear1_w)), linear2_w)
        return e.view(-1, self.rel_dim)


class BilinearEnergy(nn.Module):
    def __init__(self, rel_dim):
        super().__init__()
        self.n_bilinear = nn.Bilinear(rel_dim, rel_dim, 1)

    def forward(self, ue, ve):
        return self.n_bilinear(ue, ve)


class SynLossModule(nn.Module):
    def __init__(self, unk, semvec, synvec_model=None, batch_size=65536, eps=1e-8):
        super(SynLossModule, self).__init__()
        self.n_synvec = synvec_model if isinstance(synvec_model, MiniSynVecModel) else MiniSynVecModel.get_default()
        self.n_synvec.eval()
        self.embed_dim = self.n_synvec.word_dim
        self.rel_tot = self.n_synvec.rel_tot
        self.batch_size = batch_size
        self.splen = unk + 3
        self.n_emb = PartialTrainableEmbedding(semvec, self.embed_dim, self.splen) \
            .to(self.n_synvec.n_root_v.device)
        self.eps = eps

        self.unk_head_map = Parameter(torch.ones(self.n_emb.num_embeddings, dtype=torch.float), requires_grad=False)
        self.root_map_cache = None

    def forward(self, samples):
        h, t, r = samples[:, 0], samples[:, 1], samples[:, 2]
        h = self.n_emb(h)
        t = self.n_emb(t)
        return self.n_synvec(h, t, r)

    def batch_belief(self, samples):
        samples = torch.LongTensor(samples).to(self.n_synvec.n_root_v.device)
        res_mat = None
        for idx in range(0, len(samples), self.batch_size):
            t_s = samples[idx:idx + self.batch_size]
            t_s = self(t_s).detach()
            if res_mat is None:
                res_mat = t_s
            else:
                res_mat = torch.cat([res_mat, t_s])
        # print(len(res_mat))
        # res_mat = torch.tensor(res_mat).squeeze()
        return res_mat.squeeze()

    @functools.lru_cache(2048)
    def tail_map(self, h, r, fillsp=True, unk=0.0, log=True):
        if h >= self.splen:
            samples = []
            for t in range(self.n_emb.num_embeddings):
                samples.append((h, t, r))
            res = self.batch_belief(samples)
        else:
            res = self.unk_head_map.clone()
        if fillsp:
            res[:self.splen] = 0
            if unk != 0:
                res[3:self.splen] = unk
        if self.eps is not None:
            res = torch.clamp(res, min=self.eps)
        return (torch.log(res) if log else res).to(torch.float16)

    def root_map(self, log=True, fillsp=True):
        if self.root_map_cache is not None:
            return self.root_map_cache
        h = self.n_synvec.n_root_v.expand(self.batch_size, -1).to(self.n_synvec.n_root_v.device)
        r = torch.tensor([0]).expand(self.batch_size, -1).to(self.n_synvec.n_root_v.device)  # root
        res = None
        t_lst = torch.LongTensor(list(range(self.n_emb.num_embeddings))).to(self.n_synvec.n_root_v.device)
        for idx in range(0, len(t_lst), self.batch_size):
            s_t = t_lst[idx: idx + self.batch_size]
            s_h = h[:s_t.size(0)]
            s_r = r[:s_t.size(0)]
            t_res = self.n_synvec(s_h, self.n_emb(s_t), s_r).detach()
            if res is None:
                res = t_res
            else:
                res = torch.cat([res, t_res])
        res = res.squeeze()
        if fillsp:
            res[:self.splen] = 0
        if self.eps is not None:
            res = torch.clamp(res, min=self.eps)
        self.root_map_cache = (torch.log(res) if log else res).to(torch.float16)
        return self.root_map_cache
