import torch
from torch import nn as nn
from torch.autograd import Variable
from torch.nn import functional as F

from pysyngen.networks.common import PlainLSTM, PlainGRU, PartialTrainableEmbedding


class VariationalEncoderRNN(nn.Module):
    def __init__(self, rnn_type, input_dim,
                 hidden_dim, num_layers, z_dim,
                 dropout, variational):
        super(VariationalEncoderRNN, self).__init__()
        self.rnn_type = rnn_type
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        self.z_dim = z_dim
        self.dropout = dropout
        self.variational = variational

        if self.rnn_type == 'lstm':
            self.n_enc = PlainLSTM(self.input_dim, self.hidden_dim, self.num_layers, True, self.dropout)
        elif self.rnn_type == 'gru':
            self.n_enc = PlainGRU(self.input_dim, self.hidden_dim, self.num_layers, True, self.dropout)
        # self.n_enc_out = nn.Linear(self.hidden_dim, self.vocab_sz)
        self.n_z_mu = nn.Linear(2 * self.hidden_dim, self.z_dim)
        self.n_z_logvar = nn.Linear(2 * self.hidden_dim, self.z_dim)

    def forward(self, inputs, inputs_len, component=False):
        batch_sz = inputs.size(0)
        x, ctx = self.n_enc(inputs, inputs_len)
        x = x[torch.arange(batch_sz), inputs_len - 1]
        if self.variational:
            z, z_mu, z_logvar = self.reparameterize(x, True)
        else:
            z, z_mu, z_logvar = self.reparameterize(x, False)
        return z if component else (x, ctx, z, z_mu, z_logvar)

    def reparameterize(self, h, sample=True):
        batch_size = h.size(0)
        # have two layer lstm .
        # z = z.repeat(1, 2).view(batch_size, -1, self.latent_dim)
        z_mu = self.n_z_mu(h)
        z_logvar = 0
        if sample:
            z = Variable(torch.randn([batch_size, self.z_dim], device=h.device))
            z_logvar = self.n_z_logvar(h)
            z = z * torch.exp(0.5 * z_logvar) + z_mu
        else:
            z = z_mu
        return z, z_mu, z_logvar

    def load_from_whole_state_dict(self, vaemodel):
        prev = vaemodel.n_vae_enc.state_dict()
        self.load_state_dict(prev)

    def freeze_me(self):
        for p in self.parameters():
            p.requires_grad = False

    def extract_param(self):
        return {
            'rnn_type': self.rnn_type,
            'input_dim': self.input_dim,
            'hidden_dim': self.hidden_dim,
            'num_layers': self.num_layers,
            'z_dim': self.z_dim,
            'dropout': self.dropout,
            'variational': self.variational
        }


class VariationalAutoEncoderRNN(nn.Module):
    def __init__(self, vocab, rnn_type, embed_dim,
                 hidden_dim, num_layers, z_dim,
                 dropout,
                 semvec=None, variational=True):
        super(VariationalAutoEncoderRNN, self).__init__()
        self.vocab_sz = len(vocab)
        self.rnn_type = rnn_type
        self.embed_dim = embed_dim
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        self.z_dim = z_dim
        self.dropout = dropout
        self.variational = variational

        self.vocab = vocab
        self.bos_id = vocab['<bos>']
        self.eos_id = vocab['<eos>']
        self.pad_id = vocab['<pad>']
        current_unk_id = vocab['<unk>']
        for i in range(1, len(self.vocab) - 3):
            tmp_unk_id = vocab[f'<unk{i}>']
            if tmp_unk_id <= current_unk_id:
                break
            else:
                current_unk_id = tmp_unk_id
        self.splen = current_unk_id + 1

        # networks
        if semvec is None:
            # we use a fully trainable embedding layer as default
            self.n_emb = nn.Embedding(self.vocab_sz, self.embed_dim)
        else:
            self.n_emb = PartialTrainableEmbedding(semvec, self.embed_dim, self.splen)
        self.n_vae_enc = VariationalEncoderRNN(self.rnn_type, self.embed_dim, self.hidden_dim, self.num_layers,
                                               self.z_dim, self.dropout, self.variational)
        # self.n_ctx_bridger = nn.Linear()
        if self.rnn_type == 'lstm':
            self.n_dec = PlainLSTM(self.embed_dim + self.z_dim, self.hidden_dim, self.num_layers, False, self.dropout)
        elif self.rnn_type == 'gru':
            self.n_dec = PlainGRU(self.embed_dim + self.z_dim, self.hidden_dim, self.num_layers, False, self.dropout)
        self.n_dec_out = nn.Linear(self.hidden_dim, self.vocab_sz)

    def train_forward(self, src, src_len, tgt, tgt_len):
        batch_size = src.size(0)
        # forward
        z, z_mu, z_logvar = self.encode(src, src_len)
        x = self.decode_train(z, tgt[:, :-1], tgt_len - 1)
        # losses
        # kl loss
        if self.variational:
            loss_kl = -0.5 * torch.sum(1 + z_logvar - z_mu.pow(2) - z_logvar.exp())
            loss_kl = loss_kl / batch_size
        else:
            loss_kl = 0
        # rec loss
        tgt = tgt[:, 1:].contiguous().view(-1)
        loss_rec = torch.gather(x.view(-1, x.size(2)), 1, tgt.unsqueeze(1)).squeeze(1)
        loss_rec = - torch.sum(loss_rec * (1.0 - torch.eq(tgt, self.pad_id).float())) / (
                    tgt_len.sum().item() - batch_size)
        return loss_rec, loss_kl

    def encode(self, inputs, inputs_len, all=False):
        x = self.n_emb(inputs)
        x, ctx, z, z_mu, z_logvar = self.n_vae_enc(x, inputs_len)
        if all:
            return z, z_mu, z_logvar, x, ctx
        else:
            return z, z_mu, z_logvar

    def decode_train(self, z, tgt_seq, tgt_len):
        tgt_seq = self.n_emb(tgt_seq)
        tgt_seq = torch.cat([tgt_seq, z.unsqueeze(1).expand(-1, tgt_seq.size(1), -1)], dim=-1)
        out_seq, _ = self.n_dec(tgt_seq, tgt_len)
        out_logits = self.n_dec_out(out_seq)
        # return out_logits, out_seq, out_ctx
        # out_logits, out_seq, out_ctx = self.decode_train_step(z, tgt_seq, tgt_len)
        out_logits = F.log_softmax(out_logits, dim=-1)
        return out_logits

    def decode_greedy(self, z, max_length):
        batch_size = z.size(0)
        len_1 = torch.LongTensor([1] * batch_size)
        translations = [[] for i in range(batch_size)]
        pending = set(range(batch_size))
        prev_words = batch_size * [self.bos_id]
        out_ctx = None

        while len(pending) > 0:
            input_seq = Variable(torch.LongTensor([prev_words]).to(z.device)).transpose(0, 1)
            input_seq = self.n_emb(input_seq)
            input_seq = torch.cat([input_seq, z.unsqueeze(1)], dim=-1)
            out_seq, out_ctx = self.n_dec(input_seq, len_1, ctx=out_ctx)
            out_logits = F.log_softmax(self.n_dec_out(out_seq), dim=-1)
            prev_words = out_logits.max(dim=2)[1].squeeze().data.detach().tolist()
            for i in pending.copy():
                if prev_words[i] == self.eos_id:
                    pending.discard(i)
                else:
                    translations[i].append(prev_words[i])
                    if len(translations[i]) >= max_length:
                        pending.discard(i)
        translations = self.id2word(translations)
        return translations

    def decode_beam(self, z, beam_size, max_length):
        batch_size = z.size(0)
        z = z.repeat(beam_size, 1)
        len_1 = torch.LongTensor([1] * batch_size * beam_size)
        pending = set(range(batch_size))
        prev_words = batch_size * beam_size * [self.bos_id]
        out_ctx = None

        translations = [[] for i in range(batch_size)]
        translation_scores = batch_size * [-float('inf')]
        # arrangements: top for every sentence in batch, then no2, then no3....
        hypotheses = batch_size * [(0.0, [])] + (beam_size - 1) * batch_size * [
            (-float('inf'), [])]  # (score, translation)

        while len(pending) > 0:
            # run through decoder
            input_seq = Variable(torch.LongTensor([prev_words]).to(z.device)).transpose(0, 1)
            input_seq = self.n_emb(input_seq)
            input_seq = torch.cat([input_seq, z.unsqueeze(1)], dim=-1)
            out_seq, out_ctx = self.n_dec(input_seq, len_1, ctx=out_ctx)
            out_logits = F.log_softmax(self.n_dec_out(out_seq), dim=-1)
            # top 1 word
            prev_words = out_logits.max(dim=2)[1].squeeze().data.detach().tolist()
            # top beam words
            word_scores, words = out_logits.topk(k=beam_size + 1, dim=2, sorted=False)
            word_scores = word_scores.squeeze(1).data.detach().tolist()  # (beam_size*batch_size) * (beam_size+1)
            # print(word_scores)
            words = words.squeeze(1).data.detach().tolist()

            for sentence_idx in pending.copy():
                # deal with each sentence
                candidates = []  # (score, index, word)
                for rank in range(beam_size):
                    beam_idx = sentence_idx + rank * batch_size
                    # for each beam we are looking for beam_size+1 candidates for next words
                    # calculate their score and store them in candidates[]
                    for i in range(beam_size + 1):
                        word = words[beam_idx][i]
                        # existing score + current word score
                        score = hypotheses[beam_idx][0] + word_scores[beam_idx][i]
                        if word != self.eos_id:
                            candidates.append((score, beam_idx, word))
                        elif score > translation_scores[sentence_idx]:
                            translations[sentence_idx] = hypotheses[beam_idx][1] + [word]
                            translation_scores[sentence_idx] = score
                # beam_size * (beam_size + 1) candidates are listed, looking for top (beam_size) candidates
                best = []  # score, word, translation, hidden, output
                for score, beam_index, word in sorted(candidates, reverse=True)[:beam_size]:
                    translation = hypotheses[beam_index][1] + [word]
                    # best.append(
                    #     (score, word, translation, hidden[:, current_index, :].data, output[current_index].data))
                    best.append(
                        (score, word, translation, self.n_dec.indexing_ctx(out_ctx, beam_index)))
                for rank, (score, word, translation, h) in enumerate(best):
                    new_beam_index = sentence_idx + rank * batch_size
                    hypotheses[new_beam_index] = (score, translation)
                    prev_words[new_beam_index] = word
                    out_ctx = self.n_dec.setting_ctx(out_ctx, new_beam_index, h)
                # my top option is longer than max or my top option is worse than existing best translation
                if len(hypotheses[sentence_idx][1]) >= max_length or \
                        translation_scores[sentence_idx] > hypotheses[sentence_idx][0]:
                    pending.discard(sentence_idx)
                    if len(translations[sentence_idx]) == 0:
                        translations[sentence_idx] = hypotheses[sentence_idx][1]
                        translation_scores[sentence_idx] = hypotheses[sentence_idx][0]
        # res = [' '.join(s).replace('@@ ', '') for s in res]
        translations = self.id2word(translations)
        return translations, translation_scores

    def id2word(self, sentences):
        words = []
        for sent in sentences:
            words.append(self.vocab.lookup_tokens(sent))
        return words

    def extract_components(self):
        return {'sp_emb': self.n_emb.n_spec.state_dict() if isinstance(self.n_emb, PartialTrainableEmbedding) else None,
                'enc_param': self.n_vae_enc.extract_param(),
                'enc_state_dict': self.n_vae_enc.state_dict()}

    @staticmethod
    def test_data_text(rnn='lstm'):
        # voc = torch.load('Data/vocab/glove6B_unk1.pt')
        from torchtext.vocab import vocab
        voc = vocab(
            {'<pad>': 1, '<bos>': 1, '<eos>': 2, '<unk>': 10, 'a': 3, 'b': 4, 'c': 5, 'd': 6, 'e': 7, 'f': 8, 'g': 9})
        voc.set_default_index(3)
        src_seq = torch.tensor([[1, 4, 5, 7, 2, 0, 0, 0],
                                [1, 7, 6, 2, 0, 0, 0, 0],
                                [1, 4, 5, 7, 8, 9, 5, 2],
                                [1, 6, 6, 4, 5, 6, 2, 0]])
        src_len = torch.tensor([5, 4, 8, 7])
        tgt_seq = torch.tensor([[1, 4, 4, 4, 2, 0, 0],
                                [1, 5, 6, 4, 4, 6, 2],
                                [1, 5, 4, 8, 8, 2, 0],
                                [1, 9, 6, 5, 4, 2, 0]])
        tgt_len = torch.tensor([5, 7, 6, 6])
        mod = VariationalAutoEncoderRNN(voc, rnn, 50, 50, 2, 50, 0.1)
        return mod, voc, src_seq, src_len, tgt_seq, tgt_len
