import os
import pickle

import configargparse
import torch.autograd
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.profiler import PyTorchProfiler

from pysyngen.data.datamodule import Src2TgtLenDataModule
from pysyngen.model import TextVAEModel, SemSynGeneratorModel
from pysyngen.utils import basic_config_parser, arg_grad_acc

PLMODEL = {
    'text': TextVAEModel,
    'unsup': SemSynGeneratorModel
}


class ModelRunner:
    def __init__(self, model=None, mode='text'):
        self.mode = mode.lower()

        ModelClass = PLMODEL[self.mode]

        os.makedirs(f'TrainedModel/{self.mode}', exist_ok=True)

        if model is None:
            basic_config_parser()
            self.setup_parameters()
            ModelClass.setup_parameters()
            p = configargparse.get_argument_parser()
            self.opt = p.parse_args()
            self.opt.mode = self.mode
            self.opt.vocabname = f'{self.opt.vocab}_unk{self.opt.num_unk}'
            self.model = ModelClass(self.opt)
            self.ckpt_name = None
        else:
            self.model = ModelClass.load_from_checkpoint(model)
            self.opt = self.model.hparams
            self.ckpt_name = model

        self.data = Src2TgtLenDataModule(self.opt.dataset, self.opt.vocab, self.opt.num_unk,
                                         self.opt.maxlen, self.opt.mode, self.opt.batch_size)
        self.data.prepare_data()

        self.version_name = self.get_version_name()

        os.makedirs(f'TrainedModel/{self.mode}/{self.opt.name}/{self.version_name}', exist_ok=True)

        # if model is None:
        self.tb_logger = TensorBoardLogger(f'Logs/{self.mode}', name=self.opt.name, version=self.version_name)
        # else:
        #     self.tb_logger = TensorBoardLogger(f'Logs/Text/{self.opt.name}', name='Test', version=self.version_name)

    def get_version_name(self, prediction=False):
        version_name = f'{self.opt.dataset}_{self.model.param_name(prediction)}'
        if self.opt.ver_fix is not None:
            version_name += f'_{self.opt.ver_fix}'
        return version_name

    def train(self):
        # self.data.setup('fit')
        # torch.autograd.set_detect_anomaly(True)

        ckpt_cb = ModelCheckpoint(
            dirpath='TrainedModel/{}/{}/{}/'.format(self.mode, self.opt.name, self.version_name),
            filename='{epoch}',
            monitor='val_loss',
            verbose=False,
        )

        es_cb = EarlyStopping(
            monitor='val_loss',
            # min_delta=0.0001,
            patience=self.opt.patience,
            verbose=False
        )
        # profiler = PyTorchProfiler(profile_memory=True)

        trainer = Trainer(
            logger=self.tb_logger,
            weights_save_path='TrainedModel/{}/{}/{}/'.format(self.mode, self.opt.name, self.version_name),
            check_val_every_n_epoch=1,
            accumulate_grad_batches=self.opt.gradacc,
            gpus=self.opt.gpus,
            max_epochs=self.opt.max_epochs,
            callbacks=[ckpt_cb, es_cb],
            precision=16,
            # profiler='simple'
        )

        if self.opt.lr == 0.0:
            lr_finder = trainer.tuner.lr_find(self.model, datamodule=self.data)
            print(lr_finder.results)
            self.model.hparams.lr = lr_finder.suggestion()

        self.tb_logger.log_hyperparams(self.model.hparams)
        print(self.model.hparams)

        trainer.fit(self.model, self.data, ckpt_path=self.ckpt_name)

    def predict(self, dec_method=None):
        os.makedirs(f'Predictions/{self.opt.name}', exist_ok=True)

        if dec_method is not None:
            # update decoding method and version name
            self.model.setup_test(dec_method)
            self.version_name = self.get_version_name(True)

        tb_logger = TensorBoardLogger(f'Logs/predictions', name=self.opt.name, version=self.version_name)

        trainer = Trainer(
            logger=tb_logger,
            gpus=self.opt.gpus,
            precision=16,
            # profiler='simple'
        )

        pred_res = trainer.predict(self.model, self.data)
        res = {
            'pred': []
        }
        for batch in pred_res:
            res['pred'].extend(batch)
        res['vocab'] = self.model.hparams.vocabname
        res['dsname'] = self.data.preper.prep_ds
        fn_res = f'Predictions/{self.opt.name}/{self.mode}_{self.version_name}.res.pkl'
        with open(fn_res, 'wb') as f_res:
            pickle.dump(res, f_res)
        return fn_res

    def setup_parameters(self):
        parser = configargparse.get_argument_parser()
        trainer = parser.add_argument_group('Trainer', 'Trainer settings')
        trainer.add_argument('--gpus', default=1, type=int, help='Use GPU or not')
        trainer.add_argument('--batch_size', default=256, type=int, help='Batch size')
        trainer.add_argument('--max_epochs', default=50, type=int, help='Max number of epochs')
        trainer.add_argument('--patience', default=20, type=int, help='Patience before stop')
        trainer.add_argument('--gradacc', default=1, type=arg_grad_acc, help='Grad accumulation settings')
        data = parser.add_argument_group('Data', 'Dataset settings')
        data.add_argument('--dataset', default='SimpleWiki.v1', type=str, help='Dataset name')
        data.add_argument('--vocab', default='glove6B', type=str, help='Vocabulary name')
        data.add_argument('--num_unk', default=1, type=int, help='Number of unknown tokens')
        data.add_argument('--maxlen', default=0, type=int, help='Max length of samples')
