import argparse
import os
import pickle

import configargparse
import datasets
import numpy as np
import torch
from tqdm import tqdm


def basic_config_parser():
    parser = configargparse.get_argument_parser(ignore_unknown_config_file_keys=True)

    parser.add_argument('-c', '--config', is_config_file=True, help='Config file path')
    # general settings
    general = parser.add_argument_group('General', 'General settings')
    general.add_argument('--name', required=True, type=str, help='Model name')
    general.add_argument('--ver_fix', default=None, type=str, help='Override version name for logs and checkpoints')


def read_vectors(path, dim, pad_first=False, cache=True):  # read top n word vectors, i.e. top is 10000
    lines_num = 0
    vectors = []
    iw = []
    wi = {}
    if os.path.exists(f'{path}.pkl'):
        f_v = open(f'{path}.pkl', 'rb')
        vectors = pickle.load(f_v)
        f_iw = open(f'{path}.iw.pkl', 'rb')
        iw = pickle.load(f_iw)
        f_v.close()
        f_iw.close()
    else:
        with open(path, encoding='utf-8', errors='ignore') as f:
            for line in f:
                lines_num += 1
                tokens = line.rstrip().split(' ')
                vectors.append(np.asarray([float(x) for x in tokens[1:]]))
                iw.append(tokens[0])
        if cache:
            f_v = open(f'{path}.pkl', 'wb')
            pickle.dump(vectors, f_v)
            f_iw = open(f'{path}.iw.pkl', 'wb')
            pickle.dump(iw, f_iw)
            f_v.close()
            f_iw.close()
    if pad_first:
        if iw[0] != '<pad>':
            vectors.insert(0, np.asarray([0.0] * len(vectors[0])))
            iw.insert(0, '<pad>')
    for i, w in enumerate(iw):
        wi[w] = i
    vectors = np.array(vectors)
    return torch.Tensor(vectors)[:, :dim], iw, wi


def arg_grad_acc(arg: str):
    if arg[0] == '{' and arg[-1] == '}':
        arg = arg[1:-1]
        ret_arg = {}
        opts = arg.split(',')
        for opt in opts:
            sp = opt.split(':')
            try:
                if opt == sp[0]:
                    return int(opt)
                else:
                    ret_arg[int(sp[0])] = int(sp[1])
            except:
                print('Abnormal grad accumulation args inside the dict')
                continue
    else:
        try:
            ret_arg = int(arg)
        except:
            print('Abnormal grad accumulation args')
            ret_arg = 1
    if ret_arg == {}:
        ret_arg = 1
    return ret_arg


def arg_bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def export_dataset(dsname):
    ds = datasets.load_from_disk(f'Data/ds/{dsname}')
    os.makedirs(f'Data/export/{dsname}', exist_ok=True)
    train_cnt = 0
    test_cnt = 0

    with open(f'Data/export/{dsname}/train.source', 'w') as f_src, \
            open(f'Data/export/{dsname}/train.target', 'w') as f_tgt:
        for sample in tqdm(ds['train']):
            f_src.write(f"{' '.join(sample['source'])}\n")
            f_tgt.write(f"{' '.join(sample['target'])}\n")
            train_cnt += 1

    with open(f'Data/export/{dsname}/test.source', 'w') as f_src, \
            open(f'Data/export/{dsname}/test.target', 'w') as f_tgt:
        for sample in tqdm(ds['test']):
            f_src.write(f"{' '.join(sample['source'])}\n")
            f_tgt.write(f"{' '.join(sample['target'])}\n")
            test_cnt += 1

    print(train_cnt, len(ds['train']))
    print(test_cnt, len(ds['test']))


def tree_depth(tree):
    depth = [-1] * len(tree)

    def fill_me(i):
        # filled
        if depth[i] != -1:
            return
        # root
        if tree[i] == 0:
            depth[i] = 0
            return
        # non-root
        head = tree[i] - 1
        if depth[head] == -1:
            fill_me(head)
        depth[i] = depth[head] + 1
        return

    for idx in range(len(tree)):
        fill_me(idx)

    assert min(depth) == 0
    return depth
